"""Entrypoints to service functions through a latch_cli."""

import os
import sys
from pathlib import Path
from textwrap import dedent
from typing import Callable, List, Optional, Tuple, TypeVar, Union

import click
from packaging.version import parse as parse_version
from typing_extensions import ParamSpec

import latch_cli.click_utils
from latch.ldata._transfer.progress import Progress as _Progress
from latch_cli.click_utils import EnumChoice
from latch_cli.exceptions.handler import CrashHandler
from latch_cli.services.cp.autocomplete import complete as cp_complete
from latch_cli.services.cp.autocomplete import remote_complete
from latch_cli.services.init.init import template_flag_to_option
from latch_cli.services.local_dev import TaskSize
from latch_cli.utils import (
    AuthenticationError,
    WorkflowType,
    get_auth_header,
    get_latest_package_version,
    get_local_package_version,
)
from latch_cli.workflow_config import BaseImageOptions

latch_cli.click_utils.patch()

crash_handler = CrashHandler()

P = ParamSpec("P")
T = TypeVar("T")


def requires_login(f: Callable[P, T]) -> Callable[P, T]:
    def decorated(*args, **kwargs):
        try:
            get_auth_header()
        except AuthenticationError as e:
            click.secho(
                dedent("""
                Unable to authenticate with Latch.

                If you are on a machine with a browser, run `latch login`.

                If not, navigate to `https://console.latch.bio/settings/developer` on a different machine, select `Access Tokens`, and copy your `API Key` to `~/.latch/token` on this machine.
                If you do not see this value in the console, make sure you are logged into your personal workspace.
                """).strip("\n"),
                fg="red",
            )
            raise click.exceptions.Exit(1) from e

        return f(*args, **kwargs)

    return decorated


@click.group(
    "latch",
    context_settings={
        "max_content_width": 160,
    },
)
@click.version_option(package_name="latch")
def main():
    """
    Collection of command line tools for using the Latch SDK and
    interacting with the Latch platform.
    """
    if os.environ.get("LATCH_SKIP_VERSION_CHECK") is not None:
        return

    local_ver = parse_version(get_local_package_version())
    latest_ver = parse_version(get_latest_package_version())
    if local_ver < latest_ver:
        click.secho(
            dedent(f"""
                WARN: Your local version of latch ({local_ver}) is out of date. This may result in unexpected behavior.
                Please upgrade to the latest version ({latest_ver}) using `python3 -m pip install --upgrade latch`.
                """).strip("\n"),
            fg="yellow",
        )

    crash_handler.init()


"""
LOGIN COMMANDS
"""


@main.command("login")
@click.option(
    "--connection",
    type=str,
    default=None,
    help="Specific AuthO connection name e.g. for SSO.",
)
def login(connection: Optional[str]):
    """Manually login to Latch."""

    crash_handler.message = "Unable to log in"
    crash_handler.pkg_root = str(Path.cwd())

    from latch_cli.services.login import login

    login(connection)
    click.secho("Successfully logged into LatchBio.", fg="green")


@requires_login
@main.command("workspace")
def workspace():
    """Spawns an interactive terminal prompt allowing users to choose what workspace they want to work in."""

    crash_handler.message = "Unable to fetch workspaces"
    crash_handler.pkg_root = str(Path.cwd())

    from latch_cli.services.workspace import workspace

    workspace()


"""
WORKFLOW COMMANDS
"""


@main.command("init")
@click.argument("pkg_name", nargs=1)
@click.option(
    "--template",
    "-t",
    type=click.Choice(
        list(template_flag_to_option.keys()),
        case_sensitive=False,
    ),
)
@click.option(
    "--dockerfile",
    "-d",
    help="Create a user editable Dockerfile for this workflow.",
    is_flag=True,
    default=False,
)
@click.option(
    "--base-image",
    "-b",
    help="Which base image to use for the Dockerfile.",
    type=click.Choice(
        list(BaseImageOptions._member_names_),
        case_sensitive=False,
    ),
    default="default",
)
def init(
    pkg_name: str,
    template: Optional[str] = None,
    dockerfile: bool = False,
    base_image: str = "default",
):
    """Initialize boilerplate for local workflow code."""

    crash_handler.message = f"Unable to initialize {pkg_name}"
    crash_handler.pkg_root = str(Path.cwd())

    from latch_cli.services.init import init

    created = init(pkg_name, template, dockerfile, base_image)
    if created:
        click.secho(f"Created a latch workflow in `{pkg_name}`", fg="green")
        click.secho("Run", fg="green")
        click.secho(f"\t$ latch register {pkg_name}", fg="green")
        click.secho("To register the workflow with console.latch.bio.", fg="green")
        return

    click.secho("No workflow created.", fg="yellow")


@main.command("dockerfile")
@click.argument("pkg_root", type=click.Path(exists=True, file_okay=False))
@click.option(
    "-s",
    "--snakemake",
    is_flag=True,
    default=False,
    type=bool,
    help="Generate a Dockerfile with arguments needed for Snakemake compatability",
)
def dockerfile(pkg_root: str, snakemake: bool = False):
    """Generates a user editable dockerfile for a workflow and saves under `pkg_root/Dockerfile`.

    Visit docs.latch.bio to learn more.
    """

    crash_handler.message = "Failed to generate Dockerfile."
    crash_handler.pkg_root = pkg_root

    from latch_cli.docker_utils import generate_dockerfile

    source = Path(pkg_root)
    dest = source / "Dockerfile"
    if dest.exists() and not click.confirm(
        f"Dockerfile already exists at `{dest}`. Overwrite?"
    ):
        return
    workflow_type = WorkflowType.latchbiosdk
    if snakemake is True:
        workflow_type = WorkflowType.snakemake
    generate_dockerfile(source, dest, wf_type=workflow_type)

    click.secho(f"Successfully generated dockerfile `{dest}`", fg="green")


@main.command("generate-metadata")
@click.argument("config_file", nargs=1, type=click.Path(exists=True, path_type=Path))
@click.option(
    "--yes",
    "-y",
    is_flag=True,
    default=False,
    help=(
        "Overwrite an existing `latch_metadata/parameters.py` file without confirming."
    ),
)
@click.option(
    "--no-infer-files",
    "-I",
    is_flag=True,
    default=False,
    help="Don't parse strings with common file extensions as file parameters.",
)
@click.option(
    "--no-defaults",
    "-D",
    is_flag=True,
    default=False,
    help="Don't generate defaults for parameters.",
)
def generate_metadata(
    config_file: Path, yes: bool, no_infer_files: bool, no_defaults: bool
):
    """Generate a `latch_metadata.py` file from a Snakemake config file"""

    from latch_cli.snakemake.config.parser import generate_metadata

    generate_metadata(
        config_file,
        skip_confirmation=yes,
        infer_files=not no_infer_files,
        generate_defaults=not no_defaults,
    )


@requires_login
@main.command("develop")
@click.argument("pkg_root", nargs=1, type=click.Path(exists=True, path_type=Path))
@click.option(
    "--yes",
    "-y",
    is_flag=True,
    default=False,
    type=bool,
    help="Skip the confirmation dialog.",
)
@click.option(
    "--image",
    "-i",
    type=str,
    help="Image to use for develop session.",
)
@click.option(
    "-s",
    "--snakemake",
    is_flag=True,
    default=False,
    type=bool,
    help="Start a develop session for a Snakemake workflow.",
)
def local_development(
    pkg_root: Path,
    yes: bool,
    image: Optional[str],
    snakemake: bool,
):
    """Develop workflows "locally"

    Visit docs.latch.bio to learn more.
    """

    crash_handler.message = "Error during local development session"
    crash_handler.pkg_root = str(pkg_root)

    if os.environ.get("LATCH_DEVELOP_BETA") is not None:
        from latch_cli.services.local_dev import local_development

        local_development(
            pkg_root.resolve(),
            skip_confirm_dialog=yes,
            size=TaskSize.small_task,
            image=image,
        )
    else:
        from latch_cli.services.local_dev_old import local_development

        local_development(pkg_root.resolve(), snakemake)


@requires_login
@main.command("exec")
@click.option(
    "--execution-id", "-e", type=str, help="Optional execution ID to inspect."
)
@click.option("--egn-id", "-g", type=str, help="Optional task execution ID to inspect.")
@click.option(
    "--container-index",
    "-c",
    type=int,
    help="Optional container index to inspect (only used for Map Tasks)",
)
def execute(
    execution_id: Optional[str], egn_id: Optional[str], container_index: Optional[int]
):
    """Drops the user into an interactive shell from within a task."""

    from latch_cli.services.execute.main import exec

    exec(execution_id=execution_id, egn_id=egn_id, container_index=container_index)


@requires_login
@main.command("register")
@click.argument("pkg_root", type=click.Path(exists=True, file_okay=False))
@click.option(
    "-d",
    "--disable-auto-version",
    is_flag=True,
    default=False,
    type=bool,
    help=(
        "Whether to automatically bump the version of the workflow each time register"
        " is called."
    ),
)
@click.option(
    "--remote/--no-remote",
    is_flag=True,
    default=True,
    type=bool,
    help="Use a remote server to build workflow.",
)
@click.option(
    "--docker-progress",
    type=click.Choice(["plain", "tty", "auto"], case_sensitive=False),
    default="auto",
    help=(
        "`tty` shows only the last N lines of the build log. `plain` does no special"
        " handling. `auto` chooses `tty` when stdout is a terminal and `plain`"
        " otherwise. Equivalent to Docker's `--progress` flag."
    ),
)
@click.option(
    "-y",
    "--yes",
    is_flag=True,
    default=False,
    type=bool,
    help="Skip the confirmation dialog.",
)
@click.option(
    "--open",
    "-o",
    is_flag=True,
    default=False,
    type=bool,
    help="Open the registered workflow in the browser.",
)
@click.option(
    "--snakefile",
    type=click.Path(exists=True, dir_okay=False, path_type=Path),
    default=None,
    help="Path to a Snakefile to register.",
)
@click.option(
    "--cache-tasks/--no-cache-tasks",
    "-c/-C",
    is_flag=True,
    default=False,
    type=bool,
    help=(
        "Whether or not to cache snakemake tasks. Ignored if --snakefile is not"
        " provided."
    ),
)
def register(
    pkg_root: str,
    disable_auto_version: bool,
    remote: bool,
    docker_progress: str,
    yes: bool,
    open: bool,
    snakefile: Optional[Path],
    cache_tasks: bool,
):
    """Register local workflow code to Latch.

    Visit docs.latch.bio to learn more.
    """

    use_new_centromere = os.environ.get("LATCH_REGISTER_BETA") is not None

    crash_handler.message = "Unable to register workflow."
    crash_handler.pkg_root = pkg_root

    from latch_cli.services.register import register

    register(
        pkg_root,
        disable_auto_version=disable_auto_version,
        remote=remote,
        skip_confirmation=yes,
        open=open,
        snakefile=snakefile,
        progress_plain=(docker_progress == "auto" and not sys.stdout.isatty())
        or docker_progress == "plain",
        use_new_centromere=use_new_centromere,
        cache_tasks=cache_tasks,
    )


@requires_login
@main.command("launch")
@click.argument("params_file", nargs=1, type=click.Path(exists=True))
@click.option(
    "--version",
    default=None,
    help="The version of the workflow to launch. Defaults to latest.",
)
def launch(params_file: Path, version: Union[str, None] = None):
    """Launch a workflow using a python parameter map."""

    crash_handler.message = f"Unable to launch workflow"
    crash_handler.pkg_root = str(Path.cwd())

    from latch_cli.services.launch import launch

    wf_name = launch(params_file, version)
    if version is None:
        version = "latest"

    click.secho(
        f"Successfully launched workflow named {wf_name} with version {version}.",
        fg="green",
    )


@requires_login
@main.command("get-params")
@click.argument("wf_name", nargs=1)
@click.option(
    "--version",
    default=None,
    help="The version of the workflow. Defaults to latest.",
)
def get_params(wf_name: Union[str, None], version: Union[str, None] = None):
    """Generate a python parameter map for a workflow."""
    crash_handler.message = "Unable to generate param map for workflow"
    crash_handler.pkg_root = str(Path.cwd())

    from latch_cli.services.get_params import get_params

    get_params(wf_name, version)
    if version is None:
        version = "latest"
    click.secho(
        f"Successfully generated python param map named {wf_name}.params.py with"
        f" version {version}\n Run `latch launch {wf_name}.params.py` to launch it.",
        fg="green",
    )


@requires_login
@main.command("get-wf")
@click.option(
    "--name",
    default=None,
    help="The name of the workflow to list. Will display all versions",
)
def get_wf(name: Union[str, None] = None):
    """List workflows."""
    crash_handler.message = "Unable to get workflows"
    crash_handler.pkg_root = str(Path.cwd())

    from latch_cli.services.get import get_wf

    wfs = get_wf(name)
    id_padding, name_padding, version_padding = 0, 0, 0
    for wf in wfs:
        id, name, version = wf
        id_len, name_len, version_len = len(str(id)), len(name), len(version)
        id_padding = max(id_padding, id_len)
        name_padding = max(name_padding, name_len)
        version_padding = max(version_padding, version_len)

    # TODO(ayush): make this much better
    click.secho(
        f"ID{id_padding * ' '}\tName{name_padding * ' '}\tVersion{version_padding * ' '}"
    )
    for wf in wfs:
        click.secho(
            f"{wf[0]}{(id_padding - len(str(wf[0]))) * ' '}\t{wf[1]}{(name_padding - len(wf[1])) * ' '}\t{wf[2]}{(version_padding - len(wf[2])) * ' '}"
        )


@requires_login
@main.command("preview")
@click.argument("pkg_root", nargs=1, type=click.Path(exists=True, path_type=Path))
def preview(pkg_root: Path):
    """Creates a preview of your workflow interface."""
    crash_handler.message = f"Unable to preview inputs for {pkg_root}"
    crash_handler.pkg_root = str(pkg_root)

    from latch_cli.services.preview import preview

    preview(pkg_root)


@requires_login
@main.command("get-executions")
def get_executions():
    """Spawns an interactive terminal UI that shows all executions in a given workspace"""

    crash_handler.message = "Unable to fetch executions"

    from latch_cli.services.get_executions import get_executions

    get_executions()


"""
LDATA COMMANDS
"""


@requires_login
@main.command("cp")
@click.argument("src", shell_complete=cp_complete, nargs=-1)
@click.argument("dest", shell_complete=cp_complete, nargs=1)
@click.option(
    "--progress",
    help="Type of progress information to show while copying",
    type=EnumChoice(_Progress, case_sensitive=False),
    default="tasks",
    show_default=True,
)
@click.option(
    "--verbose",
    "-v",
    help="Print file names as they are copied",
    is_flag=True,
    default=False,
    show_default=True,
)
@click.option(
    "--no-glob",
    "-G",
    help="Don't expand globs in remote paths",
    is_flag=True,
    default=False,
    show_default=True,
)
def cp(
    src: List[str],
    dest: str,
    progress: _Progress,
    verbose: bool,
    no_glob: bool,
):
    """Copy files between Latch Data and local, or between two Latch Data locations. Behaves like `cp -R` in Unix. Directories are copied recursively. If any parents of dest do not exist, the copy will fail."""
    crash_handler.message = f"Unable to copy {src} to {dest}"
    crash_handler.pkg_root = str(Path.cwd())

    from latch_cli.services.cp.main import cp

    cp(
        src,
        dest,
        progress=progress,
        verbose=verbose,
        expand_globs=not no_glob,
    )


@requires_login
@main.command("mv")
@click.argument("src", shell_complete=remote_complete, nargs=1)
@click.argument("dest", shell_complete=remote_complete, nargs=1)
@click.option(
    "--no-glob",
    "-G",
    help="Don't expand globs in remote paths",
    is_flag=True,
    default=False,
    show_default=True,
)
def mv(src: str, dest: str, no_glob: bool):
    """Move remote files in LatchData."""

    crash_handler.message = f"Unable to move {src} to {dest}"
    crash_handler.pkg_root = str(Path.cwd())

    from latch_cli.services.move import move

    move(src, dest, no_glob=no_glob)


@requires_login
@main.command("ls")
@click.option(
    "--group-directories-first",
    "--gdf",
    help="List directories before files.",
    is_flag=True,
    default=False,
)
@click.argument("paths", nargs=-1, shell_complete=remote_complete)
def ls(paths: Tuple[str], group_directories_first: bool):
    """
    List the contents of a Latch Data directory
    """

    crash_handler.message = f"Unable to display contents of {paths}"
    crash_handler.pkg_root = str(Path.cwd())

    from latch_cli.services.ls import ls

    # If the user doesn't provide any arguments, default to root
    if len(paths) == 0:
        paths = ("/",)

    for path in paths:
        if len(paths) > 1:
            click.echo(f"{path}:")

        ls(
            path,
            group_directories_first=group_directories_first,
        )

        if len(paths) > 1:
            click.echo("")


@requires_login
@main.command("rmr")
@click.argument("remote_path", nargs=1, type=str)
@click.option(
    "-y",
    "--yes",
    is_flag=True,
    default=False,
    type=bool,
    help="Skip the confirmation dialog.",
)
@click.option(
    "--no-glob",
    "-G",
    help="Don't expand globs in remote paths",
    is_flag=True,
    default=False,
    show_default=True,
)
@click.option(
    "--verbose",
    "-v",
    help="Print all files when deleting",
    is_flag=True,
    default=False,
    show_default=True,
)
def rmr(remote_path: str, yes: bool, no_glob: bool, verbose: bool):
    """Deletes a remote entity."""
    crash_handler.message = f"Unable to delete {remote_path}"
    crash_handler.pkg_root = str(Path.cwd())

    from latch_cli.services.rm import rmr

    rmr(remote_path, skip_confirmation=yes, no_glob=no_glob, verbose=verbose)


@requires_login
@main.command("mkdirp")
@click.argument("remote_directory", nargs=1, type=str)
def mkdir(remote_directory: str):
    """Creates a new remote directory."""
    crash_handler.message = f"Unable to create directory {remote_directory}"
    crash_handler.pkg_root = str(Path.cwd())

    from latch_cli.services.mkdir import mkdirp

    mkdirp(remote_directory)


@requires_login
@main.command("sync")
@click.argument("srcs", nargs=-1)
@click.argument("dst", nargs=1)
@click.option(
    "--delete",
    help="Delete extraneous files from destination.",
    is_flag=True,
    default=False,
)
@click.option(
    "--ignore-unsyncable",
    help=(
        "Synchronize even if some source paths do not exist or refer to special files."
    ),
    is_flag=True,
    default=False,
)
def sync(srcs: List[str], dst: str, delete: bool, ignore_unsyncable: bool):
    """
    Update the contents of a remote directory with local data.
    """
    from latch_cli.services.sync import sync

    # todo(maximsmol): remote -> local
    # todo(maximsmol): remote -> remote
    sync(
        srcs,
        dst,
        delete=delete,
        ignore_unsyncable=ignore_unsyncable,
    )


"""
POD COMMANDS
"""


@main.group()
def pods():
    """Manage pods"""
    pass


@pods.command("stop")
@click.argument("pod_id", nargs=1, type=int, required=False)
@requires_login
def stop_pod(pod_id: Optional[int] = None):
    """Stops a pod given a pod_id or the pod from which the command is run"""
    crash_handler.message = "Unable to stop pod"

    from latch_cli.services.stop_pod import stop_pod

    if pod_id is None:
        id_path = Path("/root/.latch/id")

        try:
            pod_id = int(id_path.read_text().strip("\n"))
        except Exception as e:
            if isinstance(e, FileNotFoundError):
                err_str = f"Pod ID not found at `{id_path}`"
            elif isinstance(e, ValueError):
                err_str = f"Could not parse Pod ID at `{id_path}`"
            else:
                err_str = f"Error reading Pod ID from `{id_path}`"

            click.secho(
                f"{err_str} -- please provide a Pod ID as a command line argument.",
                fg="red",
            )
            return

    stop_pod(pod_id)


"""
TEST DATA COMMANDS
"""


@main.group(invoke_without_command=True)
@click.version_option(package_name="latch")
@click.pass_context
def test_data(ctx: click.Context):
    """Subcommands to upload and delete test data objects."""
    if ctx.invoked_subcommand is None:
        click.secho(f"{ctx.get_help()}")


@requires_login
@test_data.command("upload")
@click.argument("src_path", nargs=1, type=click.Path(exists=True))
@click.option(
    "--dont-confirm-overwrite",
    "-d",
    is_flag=True,
    default=False,
    type=bool,
    help="Automatically overwrite any files without asking for confirmation.",
)
def test_data_upload(src_path: str, dont_confirm_overwrite: bool):
    """Upload test data object."""

    crash_handler.message = f"Unable to upload {src_path} to managed bucket"
    crash_handler.pkg_root = str(Path.cwd())

    from latch_cli.services.test_data.upload import upload

    s3_url = upload(src_path, dont_confirm_overwrite)
    click.secho(f"Successfully uploaded to {s3_url}", fg="green")


@requires_login
@test_data.command("remove")
@click.argument("object_url", nargs=1, type=str)
def test_data_remove(object_url: str):
    """Remove test data object."""

    crash_handler.message = f"Unable to remove {object_url} from managed bucket"
    crash_handler.pkg_root = str(Path.cwd())

    from latch_cli.services.test_data.remove import remove

    remove(object_url)
    click.secho(f"Successfully removed {object_url}", fg="green")


@requires_login
@test_data.command("ls")
def test_data_ls():
    """List test data objects."""

    crash_handler.message = f"Unable to list objects within managed bucket"
    crash_handler.pkg_root = str(Path.cwd())

    from latch_cli.services.test_data.ls import ls

    objects = ls()
    click.secho("Listing your managed objects by full S3 path.\n", fg="green")
    for o in objects:
        print(f"\ts3://latch-public/{o}")



import re
from pathlib import Path
from typing import Optional, Union
from urllib.parse import urlparse

import gql
from latch_sdk_gql.execute import execute


def is_valid_url(raw_url: Union[str, Path]) -> bool:
    """A valid URL (as a source or destination of a LatchFile) must:
    * contain a latch or s3 scheme
    * contain an absolute path
    """
    try:
        parsed = urlparse(str(raw_url))
    except ValueError:
        return False
    if parsed.scheme not in ("latch", "s3"):
        return False
    if parsed.path != "" and not parsed.path.startswith("/"):
        return False
    return True


is_absolute_node_path = re.compile(r"^(latch)?://(?P<node_id>\d+).node(/)?$")
old_style_path = re.compile(r"^(?:(?P<account_root>account_root)|(?P<mount>mount))")


def format_path(path: str) -> str:
    match = is_absolute_node_path.match(path)
    if match is None:
        return path

    node_id = match.group("node_id")

    data = execute(
        gql.gql("""
        query ldataGetPathQ($id: BigInt!) {
            ldataGetPath(argNodeId: $id)
            ldataOwner(argNodeId: $id)
        }
        """),
        {"id": node_id},
    )

    raw: Optional[str] = data["ldataGetPath"]
    if raw is None:
        return path

    parts = raw.split("/")

    match = old_style_path.match(raw)
    if match is None:
        return path

    if match["mount"] is not None:
        bucket = parts[1]
        key = "/".join(parts[2:])
        return f"latch://{bucket}.mount/{key}"

    owner: Optional[str] = data["ldataOwner"]
    if owner is None:
        return path

    if match["account_root"] is not None:
        key = "/".join(parts[2:])
        return f"latch://{owner}.account/{key}"

    return path



"""Service to get workflows. """

from typing import List, Optional, Tuple, Union

import requests
from latch_sdk_config.latch import config

from latch_cli.utils import current_workspace, retrieve_or_login


# TODO(ayush): rewrite this to look/be better
def get_wf(wf_name: Optional[str] = None):
    """Get a list of a workflow's versions.

    This will allow users to list all owned workflows by default. Optionally, a
    user can provide a workflow name (unique with respect to a user) to
    list all versions of the specific workflow.

    Args:
        wf_name: The name of the workflow.

    Example:
        >>> get_wf("wf.__init__.alphafold_wf")
            ID      Name                            Version
            61858   wf.__init__.alphafold_wf        v2.1.0+14
            67261   wf.__init__.alphafold_wf        v2.2.3+0
            67317   wf.__init__.alphafold_wf        v2.2.3+14
            67341   wf.__init__.alphafold_wf        v2.2.3+19
            67408   wf.__init__.alphafold_wf        v2.2.3+40
            ...
    """

    token = retrieve_or_login()
    return _list_workflow_request(token, wf_name)


def _list_workflow_request(
    token: str, wf_name: Union[None, str]
) -> List[List[Tuple[str, str, str]]]:
    """Fetch a list of workflows, potentially by name."""

    headers = {
        "Authorization": f"Bearer {token}",
    }
    data = {
        "workflow_name": wf_name,
        "ws_account_id": current_workspace(),
    }

    url = config.api.workflow.list
    response = requests.post(url, headers=headers, json=data)

    wf_interface_resp = response.json()
    return wf_interface_resp.get("wfs", [])



from dataclasses import dataclass
from datetime import date, datetime
from enum import Enum
from typing import TYPE_CHECKING, Generic, List, Literal, Tuple, Type, TypeVar, Union

from typing_extensions import TypeAlias

from latch.registry.record import Record
from latch.registry.upstream_types.types import DBType
from latch.registry.upstream_types.values import EmptyCell
from latch.types.directory import LatchDir
from latch.types.file import LatchFile

if not TYPE_CHECKING:
    try:
        from enum import StrEnum
    except ImportError:

        class StrEnum(str, Enum): ...

else:

    class StrEnum(str, Enum): ...


@dataclass(frozen=True)
class InvalidValue:
    """Registry :class:`Record` value that failed validation."""

    raw_value: str
    """User-provided string representation of the invalid value.

    May be `""` (the empty string) if the value is missing but the column is required.
    """


RegistryPrimitivePythonValue: TypeAlias = Union[
    str,
    datetime,
    date,
    int,
    float,
    LatchFile,
    LatchDir,
    None,
    bool,
]

RegistryPythonValue: TypeAlias = Union[
    RegistryPrimitivePythonValue,
    Record,
    List["RegistryPythonValue"],
]

RecordValue: TypeAlias = Union[RegistryPythonValue, EmptyCell, InvalidValue]


@dataclass(frozen=True)
class Column:
    """Registry :class:`Table` column definition.

    :meth:`Table.get_columns` is the typical way to get a :class:`Column`.
    """

    key: str
    """Unique identifier within the table. Not globally unique."""
    type: Union[Type[RegistryPythonValue], Type[Union[RegistryPythonValue, EmptyCell]]]
    """Python equivalent of the stored column type."""
    # fixme(maximsmol): deal with defaults
    # default: Union[RegistryPythonValue, EmptyCell]
    upstream_type: DBType
    """Raw column type.

    Used to convert between Python values and Registry values.
    """


RegistryEnumDefinitionArg = TypeVar("RegistryEnumDefinitionArg", bound=Tuple[str, ...])


class RegistryEnumDefinition(Generic[RegistryEnumDefinitionArg]):
    def __new__(cls, *values: str):
        return RegistryEnumDefinition[Tuple[tuple(Literal[v] for v in values)]]


LinkedRecordTypeArg = TypeVar("LinkedRecordTypeArg", bound=str)


class LinkedRecordType(Generic[LinkedRecordTypeArg]):
    def __new__(cls, id: str):
        return LinkedRecordType[Literal[id]]


RegistryPrimitivePythonType: TypeAlias = Union[
    Type[str],
    Type[int],
    Type[float],
    Type[date],
    Type[datetime],
    Type[bool],
    Type[LatchFile],
    Type[LatchDir],
    Type[StrEnum],
    Type[RegistryEnumDefinition],
    Type[LinkedRecordType],
]
RegistryPythonType: TypeAlias = Union[
    RegistryPrimitivePythonType,
    Type[List[LatchFile]],
    Type[List[LatchDir]],
    Type[List[LinkedRecordType]],
]



import asyncio
import json
import os
import signal
import subprocess
import sys
import termios
import time
import traceback
import tty
from enum import Enum
from http.client import HTTPException
from pathlib import Path
from textwrap import dedent
from typing import Callable, Optional, Tuple, Union

import aioconsole
import asyncssh
import boto3
import click
import websockets.client as client
import websockets.exceptions
from latch_sdk_config.latch import config
from watchfiles import awatch

from latch_cli.constants import docker_image_name_illegal_pat
from latch_cli.tinyrequests import post
from latch_cli.utils import (
    TemporarySSHCredentials,
    current_workspace,
    identifier_suffix_from_str,
    retrieve_or_login,
)


def _get_workflow_name(pkg_root: Path, snakemake: bool) -> str:
    if snakemake:
        import latch.types.metadata as metadata
        from latch_cli.snakemake.utils import load_snakemake_metadata

        load_snakemake_metadata(pkg_root)
        if metadata._snakemake_metadata is None:
            click.secho(f"Unable to find latch_metadata in {pkg_root}.", fg="red")
            raise click.exceptions.Exit(1)
        return f"{metadata._snakemake_metadata.name}_jit_register"
    else:
        from flytekit.core.context_manager import FlyteEntities
        from flytekit.core.workflow import PythonFunctionWorkflow

        from latch_cli.centromere.utils import _import_flyte_objects

        _import_flyte_objects([pkg_root])
        for entity in FlyteEntities.entities:
            if isinstance(entity, PythonFunctionWorkflow):
                return entity.name

        click.secho(
            dedent(f"""
            Unable to find a workflow in {pkg_root}: Check that

            1. {pkg_root} is a valid workflow directory, and
            2. Your workflow function has the @workflow decorator.
            3. Pass in the --snakemake flag if you are using snakemake.
            """).strip("\n"),
            fg="red",
        )
        raise click.exceptions.Exit(1)


def _get_latest_image(pkg_root: Path, snakemake: bool) -> str:
    click.secho("Image name: ", fg="blue", nl=False)

    ws_id = current_workspace()
    if int(ws_id) < 10:
        ws_id = f"x{ws_id}"

    wf_name = identifier_suffix_from_str(
        _get_workflow_name(pkg_root, snakemake)
    ).lower()
    wf_name = docker_image_name_illegal_pat.sub("_", wf_name)

    registry_name = f"{ws_id}_{wf_name}"

    click.secho(registry_name, italic=True, bold=True)

    resp = post(
        config.api.workflow.get_latest,
        headers={"Authorization": f"Bearer {retrieve_or_login()}"},
        json={
            "registry_name": registry_name,
            "ws_account_id": current_workspace(),
        },
    )

    try:
        resp.raise_for_status()
        click.secho("Found image. Using version ", nl=False)
        latest_version = resp.json()["version"]
        click.secho(latest_version, italic=True, bold=True)
    except HTTPException:
        click.secho("Could not find any images. Retrying with new image name.\n")
        click.secho("Image name: ", fg="blue", nl=False)

        registry_name = f"{ws_id}_{pkg_root.name}"

        click.secho(registry_name, italic=True, bold=True)

        resp = post(
            config.api.workflow.get_latest,
            headers={"Authorization": f"Bearer {retrieve_or_login()}"},
            json={
                "registry_name": registry_name,
                "ws_account_id": current_workspace(),
            },
        )

        try:
            resp.raise_for_status()
            click.secho("Found image. Using version ", nl=False)
            latest_version = resp.json()["version"]
            click.secho(latest_version, italic=True, bold=True)
        except HTTPException as e:
            click.secho("Could not find any images.\n")

            click.secho(
                dedent("""
                There was an issue getting your workflow's docker image.

                Please make sure you've registered your workflow at least once.
                """).strip("\n"),
                fg="red",
            )
            raise click.exceptions.Exit(1)

    return f"{config.dkr_repo}/{registry_name}:{latest_version}"


rsync_stop_event = asyncio.Event()


async def _watch_and_rsync(
    pkg_root: Path,
    ip: str,
    ssh_port: str,
    key_path: Path,
    startup_grace_period: int = 10,
):
    async def _rsync() -> Tuple[int, str, str]:
        # todo(aidan): set up host key rather than disabling strict host checking
        command = [
            "rsync",
            f"--rsh=ssh -p {ssh_port} -i {str(key_path)} -o StrictHostKeyChecking=no",
            "--compress",
            "--recursive",
            "--links",
            "--times",
            "--devices",
            "--specials",
            f"{str(pkg_root)}/",
            f"root@{ip}:/root/",
        ]
        process = await asyncio.subprocess.create_subprocess_exec(
            *command,
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.PIPE,
        )
        rval = await process.wait()
        stdout, stderr = await process.communicate()
        return rval, stdout.decode("utf-8"), stderr.decode("utf-8")

    # run rsync a maximum of once every 500ms
    watcher = awatch(str(pkg_root), debounce=500, step=500, stop_event=rsync_stop_event)

    start_time = time.time()
    while time.time() - start_time < startup_grace_period:
        returncode, _, _ = await _rsync()
        if returncode == 0:
            break

        await asyncio.sleep(1)

    # todo(aidan): upgrade watchfiles and remove this error suppression after the next release of watchfiles
    # https://github.com/samuelcolvin/watchfiles/issues/200
    try:
        async for _ in watcher:
            rval, _, stderr = await _rsync()
            if rval != 0:
                await aioconsole.aprint(
                    f"Error running rsync: {stderr}", use_stderr=True
                )
    except RuntimeError as e:
        if not str(e) == "Already borrowed":
            raise e


class _MessageType(Enum):
    resize = "resize"
    body = "body"
    exit = "exit"


async def _get_messages(
    ws: client.WebSocketClientProtocol,
    show_output: bool,
):
    async for message in ws:
        msg = json.loads(message)
        if msg.get("Type") == _MessageType.exit.value:
            return
        if show_output:
            await aioconsole.aprint(
                msg.get("Body"),
                end="",
                flush=True,
            )


async def _send_message(
    ws: client.WebSocketClientProtocol,
    message: Union[str, bytes],
    typ: Optional[_MessageType] = None,
):
    if isinstance(message, bytes):
        message = message.decode("utf-8")
    if typ is None:
        typ = _MessageType.body
    body = {"Body": message, "Type": typ.value}
    await ws.send(json.dumps(body))
    # yield control back to the event loop,
    # see https://github.com/aaugustin/websockets/issues/867
    await asyncio.sleep(0)


async def _send_resize_message(
    ws: client.WebSocketClientProtocol,
    term_width: int,
    term_height: int,
):
    await _send_message(
        ws,
        json.dumps(
            {
                "Width": term_width,
                "Height": term_height,
            }
        ),
        typ=_MessageType.resize,
    )


async def _shell_session(
    ws: client.WebSocketClientProtocol,
):
    old_settings_stdin = termios.tcgetattr(sys.stdin.fileno())
    tty.setraw(sys.stdin)

    loop = asyncio.get_event_loop()
    reader = asyncio.StreamReader(loop=loop)

    stdin2 = os.dup(sys.stdin.fileno())
    stdout2 = os.dup(sys.stdout.fileno())

    old_sigwinch_handler = signal.getsignal(signal.SIGWINCH)

    await loop.connect_read_pipe(
        lambda: asyncio.StreamReaderProtocol(
            reader,
            loop=loop,
        ),
        os.fdopen(stdin2),
    )

    writer_transport, writer_protocol = await loop.connect_write_pipe(
        lambda: asyncio.streams.FlowControlMixin(loop=loop),
        os.fdopen(stdout2),
    )
    writer = asyncio.streams.StreamWriter(
        writer_transport,
        writer_protocol,
        None,
        loop,
    )

    resize_event_queue = asyncio.Queue()

    def new_sigwinch_handler(signum, frame):
        if isinstance(old_sigwinch_handler, Callable):
            old_sigwinch_handler(signum, frame)

        term_width, term_height = os.get_terminal_size()
        resize_event_queue.put_nowait((term_width, term_height))

    signal.signal(signal.SIGWINCH, new_sigwinch_handler)

    # get initial terminal size and send it over
    term_width, term_height = os.get_terminal_size()
    resize_event_queue.put_nowait((term_width, term_height))

    async def resize_task():
        while True:
            (term_width, term_height) = await resize_event_queue.get()
            await _send_resize_message(ws, term_width, term_height)

    async def input_task():
        while True:
            message = await reader.read(1)
            await _send_message(ws, message)

    async def output_task():
        async for output in ws:
            obj: dict = json.loads(output)
            if obj.get("Type") == "exit":
                io_task.cancel()
                return

            message = obj.get("Body")
            if message is not None:
                if isinstance(message, str):
                    message = message.encode("utf-8")
                writer.write(message)
            await writer.drain()
            await asyncio.sleep(0)

    try:
        io_task = asyncio.gather(input_task(), output_task(), resize_task())
        await io_task
    except asyncio.CancelledError: ...
    finally:
        termios.tcsetattr(sys.stdin.fileno(), termios.TCSANOW, old_settings_stdin)
        signal.signal(signal.SIGWINCH, old_sigwinch_handler)


async def _run_local_dev_session(pkg_root: Path, snakemake: bool):
    # hit the endpoint to make sure that a workflow image exists in ecr before
    # doing anything
    image_name_tagged = _get_latest_image(pkg_root, snakemake)
    key_path = pkg_root / ".latch" / "ssh_key"

    with TemporarySSHCredentials(key_path) as ssh:
        headers = {"Authorization": f"Bearer {retrieve_or_login()}"}

        try:
            cent_resp = post(
                config.api.centromere.provision,
                headers=headers,
                json={"public_key": ssh.public_key},
            )
            cent_resp.raise_for_status()

            init_local_dev_resp = post(
                config.api.centromere.start_local_dev,
                headers=headers,
                json={
                    "ws_account_id": current_workspace(),
                    "pkg_root": pkg_root.name,
                },
            )
            init_local_dev_resp.raise_for_status()
        except Exception as e:
            raise ValueError(f"Unable to provision a remote instance: {e}")

        if init_local_dev_resp.status_code == 403:
            raise ValueError("You are not authorized to use this feature.")

        try:
            cent_data = cent_resp.json()
            centromere_ip = cent_data["ip"]
        except KeyError as e:
            raise ValueError(f"Malformed response from provision endpoint: missing {e}")

        try:
            init_local_dev_data = init_local_dev_resp.json()
            docker_info = init_local_dev_data["docker"]
            access_key = docker_info["tmp_access_key"]
            secret_key = docker_info["tmp_secret_key"]
            session_token = docker_info["tmp_session_token"]
            port = init_local_dev_data["port"]
            ssh_port = init_local_dev_data["ssh_port"]
        except KeyError as e:
            raise ValueError(
                f"Malformed response from initialization endpoint: missing {e}"
            )

        # ecr setup
        ws_id = current_workspace()
        if int(ws_id) < 10:
            ws_id = f"x{ws_id}"

        try:
            ecr_client = boto3.session.Session(
                aws_access_key_id=access_key,
                aws_secret_access_key=secret_key,
                aws_session_token=session_token,
                region_name="us-west-2",
            ).client("ecr")

            ecr_auth = ecr_client.get_authorization_token()
            docker_access_token = ecr_auth["authorizationData"][0]["authorizationToken"]

        except Exception as err:
            raise ValueError(f"unable to retrieve an ecr login token") from err

        # todo(aidan): edit known hosts file with centromere ip and address to allow strict host checking
        try:
            async for ws in client.connect(
                f"ws://{centromere_ip}:{port}/ws",
                close_timeout=0,
                extra_headers=headers,
            ):
                try:
                    await aioconsole.aprint(
                        "Successfully connected to remote instance."
                    )
                    try:
                        await aioconsole.aprint("Setting up local sync...")
                        watch_task = asyncio.create_task(
                            _watch_and_rsync(
                                pkg_root, centromere_ip, ssh_port, key_path
                            )
                        )
                        await aioconsole.aprint("Done.")

                        await _send_message(ws, docker_access_token)
                        await _send_message(ws, image_name_tagged)
                        await aioconsole.aprint(f"Pulling {image_name_tagged}... ")

                        await _get_messages(ws, show_output=False)
                        await aioconsole.aprint("Image successfully pulled.\n")

                        await _get_messages(ws, show_output=True)

                        await _shell_session(ws)
                        rsync_stop_event.set()
                        await watch_task

                    except websockets.exceptions.ConnectionClosed:
                        continue
                except Exception as e:
                    await aioconsole.aprint(traceback.format_exc())
                finally:
                    await aioconsole.aprint("Exiting local development session")
                    await ws.close()
                    close_resp = post(
                        config.api.centromere.stop_local_dev,
                        headers={"Authorization": f"Bearer {retrieve_or_login()}"},
                    )
                    close_resp.raise_for_status()
                    return
        except asyncssh.ProtocolError as e:
            if "Too many authentication failures" in str(e):
                click.secho(
                    dedent("""
                    Too many authentication failures. Try resetting your ssh-agent with

                        $ ssh-add -D

                    and trying again."""),
                    fg="red",
                )
            else:
                raise e


def local_development(pkg_root: Path, snakemake: bool):
    """Starts a REPL that allows a user to interactively run tasks to help with
    debugging during workflow development.

    In this REPL, you can run tasks or scripts and make edits to them without
    having to reregister your workflow. You can also get a shell into a
    container with the same environment as the one that the workflow runs in, to
    help debug installation issues. See the full documentation for `Local
    Development` for more info.

    Like `get_executions`, this should only be called from the CLI for best
    results.

    Args:
        pkg_root: A path that points to a valid workflow directory (see the
            docs for `register`)

    """

    # ensure that rsync is installed
    try:
        subprocess.run(
            ["rsync", "--version"],
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL,
            check=True,
        )
    except FileNotFoundError:
        raise ValueError(dedent("""
                rsync is required for latch develop. Please install rsync and try again
                    linux: apt install rsync
                    mac: brew install rsync
                """))

    asyncio.run(_run_local_dev_session(pkg_root, snakemake))



import os
import re
from functools import lru_cache
from pathlib import Path
from typing import List

import click
import click.shell_completion as sc

from latch_cli.services.cp.utils import (
    _get_immediate_children_of_node,
    _get_known_domains_for_account,
)

cache = lru_cache(maxsize=None)
completion_type = re.compile(
    r"""
    ^
    (latch)?
    :/?/?
    (?P<domain>[^/]*)
    (
        (?P<parent>(/[^/]*)*)?
        (?P<path>/[^/]*)
    )?
    $
    """,
    re.VERBOSE,
)


def complete(
    ctx: click.Context,
    param: click.Argument,
    incomplete: str,
    allow_local: bool = True,
) -> List[sc.CompletionItem]:
    match = completion_type.match(incomplete)

    if match is None:
        if not allow_local:
            return []

        return _complete_local_path(incomplete)
    elif match["path"] is None or len(match["path"]) == 0:
        return _complete_domain(match)
    else:
        return _complete_remote_path(match)


def remote_complete(
    ctx: click.Context,
    param: click.Argument,
    incomplete: str,
):
    return complete(ctx, param, incomplete, allow_local=False)


@cache
def _complete_local_path(incomplete: str) -> List[sc.CompletionItem]:
    # todo(maximsmol): bash needs this, zsh probably needs the real thing
    # return [sc.CompletionItem("", type="file")]

    if incomplete == "":
        parent = Path.cwd()
        stub = ""
    else:
        p = Path(incomplete).resolve()
        parent = p.parent
        stub = p.name

    res: List[sc.CompletionItem] = []
    for sub_path in parent.iterdir():
        if not sub_path.name.startswith(stub):
            continue

        rel_path = os.path.relpath(sub_path)
        typ = "file" if sub_path.is_file() else "dir"
        res.append(sc.CompletionItem(rel_path, type=typ))

    return res


@cache
def _complete_remote_path(match: re.Match) -> List[sc.CompletionItem]:
    domain = match["domain"]
    parent = match["parent"]
    path = match["path"][1:]

    parent = f"://{domain}{parent}"
    if match[0].startswith("latch"):
        parent = f"latch{parent}"

    parent_path = parent
    if not parent_path.startswith("latch"):
        parent_path = f"latch{parent_path}"

    children = _get_immediate_children_of_node(parent_path)

    res: List[sc.CompletionItem] = []
    for child in children:
        if not child.startswith(path):
            continue

        res.append(sc.CompletionItem(f"{parent}/{child}"))

    return res


@cache
def _complete_domain(match: re.Match) -> List[sc.CompletionItem]:
    stub = match["domain"]

    res: List[sc.CompletionItem] = []
    for d in _get_known_domains_for_account():
        x = f"://{d}/"
        if not d.startswith(stub):
            continue

        if match[0].startswith("latch"):
            x = f"latch{x}"
        res.append(sc.CompletionItem(x))

    return res



import re
from typing import List

from latch_cli.services.cp.utils import _get_immediate_children_of_node
from latch_cli.utils import urljoins


def expand_pattern(remote_path: str) -> List[str]:
    parent, pattern = tuple(remote_path.rsplit("/", 1))

    if "*" not in pattern:
        return [remote_path]

    pattern_re = re.compile(re.escape(pattern).replace(r"\*", r".*"))

    children = _get_immediate_children_of_node(parent)

    res: List[str] = []
    for child in children:
        if not pattern_re.match(child):
            continue

        res.append(urljoins(parent, child))

    return res



from dataclasses import dataclass
from enum import Enum
from typing import List, Optional, Union

from dataclasses_json import dataclass_json
from flytekit.core.launch_plan import reference_launch_plan

from ..types.directory import LatchDir
from ..types.file import LatchFile


@dataclass_json
@dataclass
class SingleEndReads:
    r1: LatchFile


@dataclass_json
@dataclass
class PairedEndReads:
    r1: LatchFile
    r2: LatchFile


class ReadType(Enum):
    single = "single"
    paired = "paired"


class Strandedness(Enum):
    auto = "auto"


Replicate = Union[SingleEndReads, PairedEndReads]


@dataclass_json
@dataclass
class Sample:
    name: str
    strandedness: Strandedness
    replicates: List[Replicate]


class LatchGenome(Enum):
    RefSeq_hg38_p14 = "Homo sapiens (RefSeq hg38.p14)"
    RefSeq_T2T_CHM13v2_0 = "Homo sapiens (RefSeq T2T-CHM13v2.0)"
    RefSeq_R64 = "Saccharomyces cerevisiae (RefSeq R64)"
    RefSeq_GRCm39 = "Mus musculus (RefSeq GRCm39)"


class AlignmentTools(Enum):
    star_salmon = "Traditional Alignment + Quantification"
    salmon = "Selective Alignment + Quantification"


@reference_launch_plan(
    project="1",
    domain="development",
    name="wf.__init__.rnaseq",
    version="0.0.317-f53c0e",
)
def rnaseq(
    samples: List[Sample],
    alignment_quantification_tools: AlignmentTools,
    ta_ref_genome_fork: str,
    sa_ref_genome_fork: str,
    output_location_fork: str,
    run_name: str,
    latch_genome: LatchGenome,
    bams: List[List[LatchFile]],
    custom_gtf: Optional[LatchFile] = None,
    custom_ref_genome: Optional[LatchFile] = None,
    custom_ref_trans: Optional[LatchFile] = None,
    star_index: Optional[LatchFile] = None,
    salmon_index: Optional[LatchFile] = None,
    save_indices: bool = False,
    custom_output_dir: Optional[LatchDir] = None,
) -> List[LatchFile]: ...



from enum import Enum
from typing import Optional

from flytekit.core.launch_plan import reference_launch_plan

from ..types.directory import LatchDir
from ..types.file import LatchFile


class AlignmentMode(Enum):
    linsi = "L-INS-i"
    fftns2 = "FFT-NS-2"
    auto = "auto"


@reference_launch_plan(
    project="1656",
    domain="development",
    name="wf.__init__.mafft",
    version="1.0.0-e1a4bb",
)
def mafft(
    output_directory: Optional[LatchDir],
    unaligned_seqs: LatchFile,
    alignment_mode: AlignmentMode = AlignmentMode.auto,
    gap_penalty: float = 1.53,
    offset: float = 0.0,
    maxiterate: int = 0,
    output_file: str = "aligned_mafft.fa",
) -> LatchFile: ...



from multiprocessing.managers import SyncManager
from typing import Type

from .progress import ProgressBars
from .throttle import Throttle


class TransferStateManager(SyncManager):
    ProgressBars: Type[ProgressBars]
    Throttle: Type[Throttle]


TransferStateManager.register("ProgressBars", ProgressBars)
TransferStateManager.register("Throttle", Throttle)



"""
Minimal template workflow to show how to run workflows that use docker in Latch

For a more comprehensive template, see the assemble_and_sort workflow
For examples on how to use docker in Latch, see https://docs.latch.bio/examples/workflows_examples.html
"""

from wf.task import blastp_task

from latch.resources.launch_plan import LaunchPlan
from latch.resources.workflow import workflow
from latch.types.directory import LatchOutputDir
from latch.types.file import LatchFile
from latch.types.metadata import LatchAuthor, LatchMetadata, LatchParameter, LatchRule

metadata = LatchMetadata(
    display_name="biocontainers/blastp",
    author=LatchAuthor(
        name="biocontainers",
    ),
    parameters={
        "query_file": LatchParameter(
            display_name="Input FASTA File",
            batch_table_column=True,  # Show this parameter in batched mode.
            rules=[
                # validate the input file using regex
                LatchRule(
                    regex="(.fasta|.fa|)$",
                    message="Only fasta, fa, extensions are valid",
                )
            ],
        ),
        "database": LatchParameter(
            display_name="Database to run blastp against (.faa file)",
            batch_table_column=True,  # Show this parameter in batched mode.
        ),
        "output_directory": LatchParameter(
            display_name="Output Directory",
            batch_table_column=True,  # Show this parameter in batched mode.
        ),
    },
    tags=[],
)


@workflow(metadata)
def blast_wf(
    query_file: LatchFile, database: LatchFile, output_directory: LatchOutputDir
) -> LatchFile:
    return blastp_task(
        query_fasta=query_file, database=database, output_directory=output_directory
    )


LaunchPlan(
    blast_wf,
    "Test Data",
    {
        "query_file": LatchFile(
            "s3://latch-public/test-data/1534/host-data/P04156.fasta"
        ),
        "database": LatchFile(
            "s3://latch-public/test-data/1534/host-data/zebrafish.1.protein.faa"
        ),
        "output_directory": LatchOutputDir("latch:///biocontainers_blastp_output"),
    },
)



import os
import time
from typing import Any, Dict, Optional

from gql.gql import DocumentNode
from gql.transport.exceptions import TransportClosed, TransportServerError
from latch_sdk_gql import JsonValue
from latch_sdk_gql.execute import execute

from latch_cli import tinyrequests


# todo(rahul): move this function into latch_sdk_gql.execute
def query_with_retry(
    document: DocumentNode,
    variables: Optional[Dict[str, JsonValue]] = None,
    *,
    num_retries: int = 3,
) -> Dict[str, Any]:
    """
    Send GraphQL query request. Retry on Server or Connection failures.
    Implements exponential backoff between retries
    """
    attempt = 0
    while attempt < num_retries:
        attempt += 1
        try:
            data = execute(document, variables)
            return data
        except (TransportServerError, TransportClosed) as e:
            err = e

        if attempt < num_retries:
            # todo(rahul): tune the sleep interval based on the startup time of the vacuole
            time.sleep(2**attempt * 5)

    raise err


def get_max_workers() -> int:
    try:
        max_workers = len(os.sched_getaffinity(0)) * 4
    except AttributeError:
        cpu = os.cpu_count()
        if cpu is not None:
            max_workers = cpu * 4
        else:
            max_workers = 16

    return min(max_workers, 16)



import asyncio
import base64
import json
import os
import signal
import sys
import termios
import tty
from typing import Generic, Literal, Optional, Tuple, TypedDict, TypeVar, Union
from urllib.parse import urljoin, urlparse

import websockets.client as websockets
from latch_sdk_config.latch import NUCLEUS_URL
from typing_extensions import TypeAlias

from latch_cli.services.execute.utils import (
    ContainerNode,
    EGNNode,
    ExecutionInfoNode,
    get_container_info,
    get_egn_info,
    get_execution_info,
)
from latch_cli.utils import get_auth_header


class StdoutResponse(TypedDict):
    stream: Union[Literal["stdout"], Literal["stderr"]]
    data: str


class CloseResponse(TypedDict):
    stream: Literal["close"]


Response: TypeAlias = Union[StdoutResponse, CloseResponse]


async def get_stdio_streams():
    loop = asyncio.get_event_loop()
    reader = asyncio.StreamReader(loop=loop)

    stdin2 = os.dup(sys.stdin.fileno())
    stdout2 = os.dup(sys.stdout.fileno())

    await loop.connect_read_pipe(
        lambda: asyncio.StreamReaderProtocol(
            reader,
            loop=loop,
        ),
        os.fdopen(stdin2),
    )

    writer_transport, writer_protocol = await loop.connect_write_pipe(
        lambda: asyncio.streams.FlowControlMixin(loop=loop),
        os.fdopen(stdout2),
    )
    writer = asyncio.streams.StreamWriter(
        writer_transport,
        writer_protocol,
        None,
        loop,
    )

    return reader, writer


async def shutdown_on_close(ws: websockets.WebSocketClientProtocol, task: asyncio.Task):
    await ws.wait_closed()
    task.cancel()


async def pipe_to_remote_stdin(
    ws: websockets.WebSocketClientProtocol, reader: asyncio.StreamReader
):
    while True:
        message = await reader.read(99999)
        if message == b"":
            return

        await ws.send(
            json.dumps({
                "stream": "stdin",
                "data": base64.b64encode(message).decode("ascii"),
            })
        )


async def pipe_from_remote_stdout(
    ws: websockets.WebSocketClientProtocol, writer: asyncio.StreamWriter
):
    async for output in ws:
        if isinstance(output, bytes):
            output = output.decode()

        try:
            res = json.loads(output)
            # todo(ayush): write stderr messages to stderr lol
            if res["stream"] == "close":
                break

            writer.write(base64.b64decode(res["data"].encode("ascii")))
        except (json.JSONDecodeError, KeyError) as e:
            # todo(ayush) error surfacing
            break


async def handle_resize(resize_event_queue: asyncio.Queue):
    await resize_event_queue.put(os.get_terminal_size())


async def propagate_resize_events(
    ws: websockets.WebSocketClientProtocol,
    resize_event_queue: asyncio.Queue,
):
    while True:
        cols, rows = await resize_event_queue.get()

        await ws.send(
            json.dumps({
                "stream": "resize",
                "size": {
                    "Width": cols,
                    "Height": rows,
                },
            })
        )


async def connect(egn_info: EGNNode, container_info: Optional[ContainerNode]):
    loop = asyncio.get_event_loop()

    resize_queue: asyncio.Queue = asyncio.Queue()
    await resize_queue.put(os.get_terminal_size())

    loop.add_signal_handler(
        signal.SIGWINCH,
        lambda: asyncio.create_task(handle_resize(resize_queue)),
    )

    local_stdin, local_stdout = await get_stdio_streams()

    async with websockets.connect(
        urlparse(urljoin(NUCLEUS_URL, "/workflows/cli/shell"))
        ._replace(scheme="wss")
        .geturl(),
        close_timeout=0,
        extra_headers={"Authorization": get_auth_header()},
    ) as ws:
        request = {
            "egn_id": egn_info["id"],
            "container_index": (
                container_info["index"] if container_info is not None else None
            ),
        }

        await ws.send(json.dumps(request))

        # ayush: can't use TaskGroups bc only supported on >= 3.11
        try:
            _, pending = await asyncio.wait(
                [
                    asyncio.create_task(pipe_from_remote_stdout(ws, local_stdout)),
                    asyncio.create_task(pipe_to_remote_stdin(ws, local_stdin)),
                    asyncio.create_task(propagate_resize_events(ws, resize_queue)),
                ],
                return_when=asyncio.FIRST_COMPLETED,
            )

            for unfinished in pending:
                unfinished.cancel()

        except asyncio.CancelledError:
            pass


def exec(
    execution_id: Optional[str] = None,
    egn_id: Optional[str] = None,
    container_index: Optional[int] = None,
):
    execution_info: Optional[ExecutionInfoNode] = None
    if egn_id is None:
        execution_info = get_execution_info(execution_id)

    egn_info = get_egn_info(execution_info, egn_id)
    container_info = get_container_info(egn_info, container_index)

    old_settings_stdin = termios.tcgetattr(sys.stdin.fileno())
    tty.setraw(sys.stdin)

    try:
        asyncio.run(connect(egn_info, container_info))
    finally:
        termios.tcsetattr(sys.stdin.fileno(), termios.TCSANOW, old_settings_stdin)



from __future__ import annotations  # avoid circular type imports

from dataclasses import dataclass, field
from typing import (
    TYPE_CHECKING,
    Dict,
    List,
    Literal,
    Optional,
    Tuple,
    TypedDict,
    overload,
)

import gql
from latch_sdk_gql.execute import execute

from latch.registry.upstream_types.types import DBType
from latch.registry.upstream_types.values import DBValue

if TYPE_CHECKING:  # avoid circular type imports
    from latch.registry.types import Column, RecordValue


class NoSuchColumnError(KeyError):
    """Unknown column referenced by Registry method.

    Reloading the containing table might help.

    Attributes:
        key: The unknown column key.
    """

    def __init__(self, key: str):
        super().__init__(f"no such column: {key}")

        self.key = key


class _ColumnDefinition(TypedDict("_ColumnDefinitionReserved", {"def": DBValue})):
    key: str
    type: DBType


class _ColumnDefinitionConnection(TypedDict):
    nodes: List[_ColumnDefinition]


class _CatalogExperiment(TypedDict):
    id: str
    catalogExperimentColumnDefinitionsByExperimentId: _ColumnDefinitionConnection


class _ColumnDataNode(TypedDict):
    key: str
    data: DBValue


class _ColumnDataConnection(TypedDict):
    nodes: List[_ColumnDataNode]


class _CatalogSample(TypedDict):
    id: str
    name: str
    experiment: _CatalogExperiment
    catalogSampleColumnDataBySampleId: _ColumnDataConnection


@dataclass
class _Cache:
    """Internal cache class to organize information for a `Record`."""

    table_id: Optional[str] = None
    name: Optional[str] = None
    columns: Optional[Dict[str, Column]] = None
    values: Optional[Dict[str, RecordValue]] = None


@dataclass(frozen=True)
class Record:
    """Registry record.

    :meth:`~latch.registry.table.Table.list_records` is the typical way to get a :class:`Record`.
    """

    _cache: _Cache = field(
        default_factory=lambda: _Cache(),
        init=False,
        repr=False,
        hash=False,
        compare=False,
    )

    id: str
    """Unique identifier."""

    def load(self) -> None:
        """(Re-)populate this record instance's cache.

        Future calls to most getters will return immediately without making a network request.

        Always makes a network request.
        """
        # avoid circular type imports
        from latch.registry.types import Column, InvalidValue
        from latch.registry.utils import to_python_literal, to_python_type

        data: _CatalogSample = execute(
            gql.gql("""
            query RecordQuery($id: BigInt!) {
                catalogSample(id: $id) {
                    id
                    name
                    catalogSampleColumnDataBySampleId {
                        nodes {
                            key
                            data
                        }
                    }
                    experiment {
                        id
                        catalogExperimentColumnDefinitionsByExperimentId {
                            nodes {
                                type
                                key
                                def
                            }
                        }
                    }
                }
            }
            """),
            {"id": self.id},
        )["catalogSample"]
        # todo(maximsmol): deal with nonexistent records

        self._cache.table_id = data["experiment"]["id"]
        self._cache.name = data["name"]

        typeNodes = data["experiment"][
            "catalogExperimentColumnDefinitionsByExperimentId"
        ]["nodes"]
        # fixme(maximsmol): deal with defaults
        self._cache.columns = {
            n["key"]: Column(n["key"], to_python_type(n["type"]["type"]), n["type"])
            for n in typeNodes
        }

        valNodes = data["catalogSampleColumnDataBySampleId"]["nodes"]
        colVals = {n["key"]: n["data"] for n in valNodes}

        vals: Dict[str, RecordValue] = {}
        for k, v in colVals.items():
            col = self._cache.columns.get(k)
            if col is None:
                raise NoSuchColumnError(k)

            # todo(maximsmol): allow creating records with mismatching types
            vals[k] = to_python_literal(v, col.upstream_type["type"])

        for k, c in self._cache.columns.items():
            if k in vals:
                continue

            if not c.upstream_type["allowEmpty"]:
                vals[k] = InvalidValue("")

            # prevent keyerrors when accessing columns that don't have a value
            vals[k] = None

        self._cache.values = vals

    # get_table_id

    @overload
    def get_table_id(self, *, load_if_missing: Literal[True] = True) -> str: ...

    @overload
    def get_table_id(self, *, load_if_missing: bool) -> Optional[str]: ...

    def get_table_id(self, *, load_if_missing: bool = True) -> Optional[str]:
        """Get the ID of the table that contains this record.

        Args:
            load_if_missing:
                If true, :meth:`load` the table ID if not in cache.
                If false, return `None` if not in cache.

        Returns:
            ID of the :class:`Table` containing this record.
        """
        if self._cache.table_id is None:
            if not load_if_missing:
                return None

            self.load()

        return self._cache.table_id

    # get_name

    @overload
    def get_name(self, *, load_if_missing: Literal[True] = True) -> str: ...

    @overload
    def get_name(self, *, load_if_missing: bool) -> Optional[str]: ...

    def get_name(self, *, load_if_missing: bool = True) -> Optional[str]:
        """Get the name of this record.

        Names are unique within a table. Names are not globally unique.
        Use :attr:`id` if a globally unique identifier is required.

        Args:
            load_if_missing:
                If true, :meth:`load` the name if not in cache.
                If false, return `None` if not in cache.

        Returns:
            Name of this record.
        """
        if self._cache.name is None and load_if_missing:
            self.load()

        return self._cache.name

    # get_columns

    @overload
    def get_columns(
        self,
        *,
        load_if_missing: Literal[True] = True,
    ) -> Dict[str, Column]: ...

    @overload
    def get_columns(
        self,
        *,
        load_if_missing: bool,
    ) -> Optional[Dict[str, Column]]: ...

    def get_columns(
        self,
        *,
        load_if_missing: bool = True,
    ) -> Optional[Dict[str, Column]]:
        """Get the columns of this record's table.

        Args:
            load_if_missing:
                If true, :meth:`load` the column list if not in cache.
                If false, return `None` if not in cache.

        Returns:
            Mapping between column keys and :class:`columns <latch.registry.types.Column>`.
        """
        if self._cache.columns is None and load_if_missing:
            self.load()

        return self._cache.columns

    # get_values

    @overload
    def get_values(
        self,
        *,
        load_if_missing: Literal[True] = True,
    ) -> Dict[str, RecordValue]: ...

    @overload
    def get_values(
        self,
        *,
        load_if_missing: bool,
    ) -> Optional[Dict[str, RecordValue]]: ...

    def get_values(
        self,
        *,
        load_if_missing: bool = True,
    ) -> Optional[Dict[str, RecordValue]]:
        """Get this record's values.

        The resulting dictionary is shared between all calls to :meth:`get_values`.
        Make deep copies if independent mutation is desired.

        Args:
            load_if_missing:
                If true, :meth:`load` the values if not in cache.
                If false, return `None` if not in cache.

        Returns:
            Mapping between column keys and the corresponding value.
        """
        if self._cache.values is None and load_if_missing:
            self.load()

        return self._cache.values

    def _repr_parts(self) -> Tuple[str, Optional[Dict[str, RecordValue]]]:
        name = self.get_name(load_if_missing=False)
        if name is None:
            base = f"Record(id={self.id})"
        else:
            base = f"Record(id={self.id}, name={name})"

        values = self.get_values(load_if_missing=False)
        return base, values

    def __repr__(self):
        base, vals = self._repr_parts()

        if vals is None:
            return base

        return base + repr(vals)

    def __str__(self):
        base, vals = self._repr_parts()

        if vals is None:
            return base

        return base + str(vals)



import json
from dataclasses import asdict, dataclass
from datetime import datetime
from enum import Enum
from pathlib import Path

from pkg_resources import get_distribution

from latch_cli.constants import latch_constants


class BaseImageOptions(str, Enum):
    default = "default"
    cuda = "cuda"
    opencl = "opencl"


@dataclass(frozen=True)
class LatchWorkflowConfig:
    """Configuration for a Latch workflow."""

    latch_version: str
    """Version of the Latch SDK used to initialize the workflow"""
    base_image: str
    """Exact version of the included workflow base image"""
    date: str
    """Timestamp of the `latch init` call"""


def create_and_write_config(
    pkg_root: Path, base_image_type: BaseImageOptions = BaseImageOptions.default
):
    base_image = latch_constants.base_image

    if base_image_type != BaseImageOptions.default:
        base_image = base_image.replace(
            "latch-base", f"latch-base-{base_image_type.name}"
        )

    config = LatchWorkflowConfig(
        latch_version=get_distribution("latch").version,
        base_image=base_image,
        date=datetime.now().isoformat(),
    )

    (pkg_root / ".latch").mkdir(exist_ok=True)

    with (pkg_root / latch_constants.pkg_config).open("w") as f:
        f.write(json.dumps(asdict(config)))



import json
import os
import sys
import textwrap
import traceback
from pathlib import Path
from textwrap import dedent
from typing import Any, Dict, List, Optional, Set, Union, get_args

import click
from flyteidl.admin.launch_plan_pb2 import LaunchPlan as _idl_admin_LaunchPlan
from flyteidl.admin.task_pb2 import TaskSpec as _idl_admin_TaskSpec
from flyteidl.admin.workflow_pb2 import WorkflowSpec as _idl_admin_WorkflowSpec
from flytekit import LaunchPlan
from flytekit.configuration import Image, ImageConfig, SerializationSettings
from flytekit.models import launch_plan as launch_plan_models
from flytekit.models import literals as literals_models
from flytekit.models import task as task_models
from flytekit.models.admin import workflow as admin_workflow_models
from flytekit.tools.serialize_helpers import persist_registrable_entities
from google.protobuf.json_format import MessageToJson
from snakemake.dag import DAG
from snakemake.persistence import Persistence
from snakemake.rules import Rule
from snakemake.workflow import Workflow, WorkflowError
from typing_extensions import Self

import latch.types.metadata as metadata

from ..services.register.utils import import_module_by_path
from .serialize_utils import (
    EntityCache,
    get_serializable_launch_plan,
    get_serializable_workflow,
)
from .utils import load_snakemake_metadata
from .workflow import JITRegisterWorkflow, SnakemakeWorkflow, interface_to_parameters

RegistrableEntity = Union[
    task_models.TaskSpec,
    launch_plan_models.LaunchPlan,
    admin_workflow_models.WorkflowSpec,
]


def should_register_with_admin(entity: RegistrableEntity) -> bool:
    return isinstance(entity, get_args(RegistrableEntity))


def get_snakemake_metadata_example(name: str) -> str:
    return dedent(f"""
        from pathlib import Path
        from latch.types.metadata import SnakemakeMetadata, SnakemakeFileParameter
        from latch.types.file import LatchFile
        from latch.types.metadata import LatchAuthor

        SnakemakeMetadata(
            display_name={repr(name)},
            author=LatchAuthor(
                name="Anonymous",
            ),
            parameters={{
                "example": SnakemakeFileParameter(
                    display_name="Example Parameter",
                    type=LatchFile,
                    path=Path("example.txt"),
                )
            }},
        )
        """).lstrip()


def ensure_snakemake_metadata_exists():
    if metadata._snakemake_metadata is None:
        click.secho(
            dedent("""
                    No `SnakemakeMetadata` object was detected in your project. This
                    object needs to be defined to register this workflow with Latch.

                    Create a file named `latch_metadata.py` with the following
                    code to get started:

                    __example__

                    Find more information at docs.latch.bio.
                    """).replace(
                "__example__", get_snakemake_metadata_example("example_name")
            ),
            bold=True,
            fg="red",
        )


# todo(maximsmol): this needs to run in a subprocess because it pollutes globals
class SnakemakeWorkflowExtractor(Workflow):
    def __init__(
        self,
        pkg_root: Path,
        snakefile: Path,
        overwrite_config: Optional[Dict[str, Any]] = None,
    ):
        assert metadata._snakemake_metadata is not None
        cores = metadata._snakemake_metadata.cores
        super().__init__(
            snakefile=snakefile, overwrite_config=overwrite_config, cores=cores
        )

        self.pkg_root = pkg_root
        self._old_cwd = ""

        if overwrite_config is not None:
            print(f"Config: {json.dumps(overwrite_config, indent=2)}")

    def extract_dag(self):
        targets: List[str] = (
            [self.default_target] if self.default_target is not None else []
        )
        target_rules: Set[Rule] = set(
            self._rules[x] for x in targets if self.is_rule(x)
        )

        target_files: Set[str] = set()
        for f in targets:
            if self.is_rule(f):
                continue

            if os.path.isabs(f) or f.startswith("root://"):
                target_files.add(f)
            else:
                target_files.add(os.path.relpath(f))

        dag = DAG(
            self,
            self.rules,
            targetfiles=target_files,
            targetrules=target_rules,
            priorityrules=set(),
            priorityfiles=set(),
        )

        try:
            self.persistence = Persistence(dag=dag)
        except AttributeError:
            self._persistence = Persistence(dag=dag)

        dag.init()
        dag.update_checkpoint_dependencies()
        dag.check_dynamic()

        return dag

    def __enter__(self) -> Self:
        self._old_cwd = os.getcwd()
        os.chdir(self.pkg_root)

        return self

    def __exit__(self, typ, value, tb):
        os.chdir(self._old_cwd)

        if typ is None:
            return False

        if not isinstance(value, WorkflowError):
            return False

        msg = str(value)
        if (
            "Workflow defines configfile config.yaml but it is not present or"
            " accessible"
            in msg
        ):
            # todo(maximsmol): print the expected config path
            traceback.print_exception(typ, value, tb)
            click.secho("\n\n\nHint: ", fg="red", bold=True, nl=False, err=True)
            click.secho("Snakemake could not find a config file", fg="red", err=True)
            sys.exit(1)

        # todo(maximsmol): handle specific errors
        # WorkflowError: Failed to open source file /Users/maximsmol/projects/latchbio/latch/test/CGI_WGS_GATK_Pipeline/Snakefiles/CGI_WGS_GATK_Pipeline/Snakefiles/calc_frag_len.smk
        # FileNotFoundError: [Errno 2] No such file or directory: '/Users/maximsmol/projects/latchbio/latch/test/CGI_WGS_GATK_Pipeline/Snakefiles/CGI_WGS_GATK_Pipeline/Snakefiles/calc_frag_len.smk'
        raise RuntimeError("invalid Snakefile") from value


def snakemake_workflow_extractor(
    pkg_root: Path,
    snakefile: Path,
    overwrite_config: Optional[Dict[str, Any]] = None,
) -> SnakemakeWorkflowExtractor:
    snakefile = snakefile.resolve()

    load_snakemake_metadata(pkg_root)

    extractor = SnakemakeWorkflowExtractor(
        pkg_root=pkg_root,
        snakefile=snakefile,
        overwrite_config=overwrite_config,
    )
    with extractor:
        extractor.include(
            snakefile,
            overwrite_default_target=True,
        )
        ensure_snakemake_metadata_exists()

    return extractor


def extract_snakemake_workflow(
    pkg_root: Path,
    snakefile: Path,
    jit_wf_version: str,
    jit_exec_display_name: str,
    local_to_remote_path_mapping: Optional[Dict[str, str]] = None,
    overwrite_config: Optional[Dict[str, Any]] = None,
    cache_tasks: bool = False,
) -> SnakemakeWorkflow:
    extractor = snakemake_workflow_extractor(pkg_root, snakefile, overwrite_config)
    with extractor:
        dag = extractor.extract_dag()
        wf = SnakemakeWorkflow(
            dag,
            jit_wf_version,
            jit_exec_display_name,
            local_to_remote_path_mapping,
            cache_tasks,
        )
        wf.compile()

    return wf


def serialize_snakemake(
    wf: SnakemakeWorkflow,
    output_dir: Path,
    image_name: str,
    dkr_repo: str,
):
    image_name_no_version, version = image_name.split(":")
    default_img = Image(
        name=image_name,
        fqn=f"{dkr_repo}/{image_name_no_version}",
        tag=version,
    )
    settings = SerializationSettings(
        image_config=ImageConfig(default_image=default_img, images=[default_img]),
    )

    registrable_entity_cache: EntityCache = {}

    spec_dir = Path("spec")
    spec_dir.mkdir(parents=True, exist_ok=True)

    wf_spec = get_serializable_workflow(wf, settings, registrable_entity_cache)
    (spec_dir / "wf.json").write_text(MessageToJson(wf_spec.to_flyte_idl()))

    parameter_map = interface_to_parameters(wf.python_interface)
    lp = LaunchPlan(
        name=wf.name,
        workflow=wf,
        parameters=parameter_map,
        fixed_inputs=literals_models.LiteralMap(literals={}),
    )
    admin_lp = get_serializable_launch_plan(lp, settings, registrable_entity_cache)

    registrable_entities = [
        x.to_flyte_idl()
        for x in list(
            filter(should_register_with_admin, list(registrable_entity_cache.values()))
        )
        + [admin_lp]
    ]
    for idx, entity in enumerate(registrable_entities):
        cur = spec_dir

        if isinstance(entity, _idl_admin_TaskSpec):
            cur = cur / "tasks" / f"{entity.template.id.name}_{idx}.json"
        elif isinstance(entity, _idl_admin_WorkflowSpec):
            cur = cur / "wfs" / f"{entity.template.id.name}_{idx}.json"
        elif isinstance(entity, _idl_admin_LaunchPlan):
            cur = cur / "lps" / f"{entity.id.name}_{idx}.json"
        else:
            click.secho(
                f"Entity is incorrect formatted {entity} - type {type(entity)}",
                fg="red",
            )
            sys.exit(-1)

        cur.parent.mkdir(parents=True, exist_ok=True)
        cur.write_text(MessageToJson(entity))

    persist_registrable_entities(registrable_entities, str(output_dir))


def serialize_jit_register_workflow(
    jit_wf: JITRegisterWorkflow,
    output_dir: str,
    image_name: str,
    dkr_repo: str,
):
    image_name_no_version, version = image_name.split(":")
    default_img = Image(
        name=image_name,
        fqn=f"{dkr_repo}/{image_name_no_version}",
        tag=version,
    )
    settings = SerializationSettings(
        image_config=ImageConfig(default_image=default_img, images=[default_img]),
    )

    registrable_entity_cache: EntityCache = {}

    get_serializable_workflow(jit_wf, settings, registrable_entity_cache)

    parameter_map = interface_to_parameters(jit_wf.python_interface)
    lp = LaunchPlan(
        name=jit_wf.name,
        workflow=jit_wf,
        parameters=parameter_map,
        fixed_inputs=literals_models.LiteralMap(literals={}),
    )
    admin_lp = get_serializable_launch_plan(lp, settings, registrable_entity_cache)

    registrable_entities = [
        x.to_flyte_idl()
        for x in list(
            filter(should_register_with_admin, list(registrable_entity_cache.values()))
        )
        + [admin_lp]
    ]

    click.secho("\nSerializing workflow entities", bold=True)
    persist_registrable_entities(registrable_entities, output_dir)


def snakefile_path_in_container(snakefile: Path, pkg_root: Path) -> str:
    return str(snakefile.resolve())[len(str(pkg_root.resolve())) + 1 :]


def generate_snakemake_entrypoint(
    wf: SnakemakeWorkflow,
    pkg_root: Path,
    snakefile: Path,
    remote_output_url: Optional[str] = None,
    overwrite_config: Optional[Dict[str, str]] = None,
):
    entrypoint_code_block = textwrap.dedent(r"""
        import os
        from pathlib import Path
        import shutil
        import subprocess
        from subprocess import CalledProcessError
        from typing import NamedTuple, Dict
        import stat
        import sys
        from dataclasses import is_dataclass, asdict
        from enum import Enum

        from flytekit.extras.persistence import LatchPersistence
        import traceback

        from latch.resources.tasks import custom_task
        from latch.types.directory import LatchDir
        from latch.types.file import LatchFile

        from latch_cli.utils import get_parameter_json_value, urljoins, check_exists_and_rename
        from latch_cli.snakemake.serialize_utils import update_mapping

        sys.stdout.reconfigure(line_buffering=True)
        sys.stderr.reconfigure(line_buffering=True)


        def si_unit(num, base: float = 1000.0):
            for unit in (" ", "k", "M", "G", "T", "P", "E", "Z"):
                if abs(num) < base:
                    return f"{num:3.1f}{unit}"
                num /= base
            return f"{num:.1f}Y"


        def file_name_and_size(x: Path):
            s = x.stat()

            if stat.S_ISDIR(s.st_mode):
                return f"{'D':>8} {x.name}/"

            return f"{si_unit(s.st_size):>7}B {x.name}"

    """).lstrip()

    entrypoint_code_block += "\n\n".join(
        task.get_fn_code(
            snakefile_path_in_container(snakefile, pkg_root),
            remote_output_url,
            overwrite_config,
        )
        for task in wf.snakemake_tasks
    )

    entrypoint = pkg_root / "latch_entrypoint.py"
    entrypoint.write_text(entrypoint_code_block + "\n")


def generate_jit_register_code(
    wf: JITRegisterWorkflow,
    pkg_root: Path,
    snakefile: Path,
    version: str,
    image_name: str,
    account_id: str,
) -> Path:
    code_block = textwrap.dedent(r"""
        import json
        import os
        import subprocess
        import tempfile
        import textwrap
        import time
        import sys
        from functools import partial
        from pathlib import Path
        import shutil
        import typing
        from typing import NamedTuple, Optional, TypedDict, Dict
        import hashlib
        from urllib.parse import urljoin
        from dataclasses import is_dataclass, asdict
        from enum import Enum

        import stat
        import base64
        import boto3
        import boto3.session
        import google.protobuf.json_format as gpjson
        import gql
        import requests
        from flytekit.core import utils
        from flytekit.extras.persistence import LatchPersistence
        from latch_cli import tinyrequests
        from latch_cli.centromere.utils import _construct_dkr_client
        from latch_sdk_config.latch import config
        from latch_cli.services.register.register import (
            _print_reg_resp,
            _recursive_list,
            register_serialized_pkg,
            print_and_write_build_logs,
            print_upload_logs,
        )
        from latch_cli.snakemake.serialize import (
            extract_snakemake_workflow,
            generate_snakemake_entrypoint,
            serialize_snakemake,
        )
        from latch_cli.utils import get_parameter_json_value, check_exists_and_rename
        import latch_cli.snakemake
        from latch_cli.snakemake.serialize_utils import update_mapping
        from latch_cli.utils import urljoins

        from latch import small_task
        from latch_sdk_gql.execute import execute
        from latch.types.directory import LatchDir
        from latch.types.file import LatchFile

        try:
            import latch_metadata.parameters as latch_metadata
        except ImportError:
            import latch_metadata

        sys.stdout.reconfigure(line_buffering=True)
        sys.stderr.reconfigure(line_buffering=True)


        def si_unit(num, base: float = 1000.0):
            for unit in (" ", "k", "M", "G", "T", "P", "E", "Z"):
                if abs(num) < base:
                    return f"{num:3.1f}{unit}"
                num /= base
            return f"{num:.1f}Y"

        def file_name_and_size(x: Path):
            s = x.stat()

            if stat.S_ISDIR(s.st_mode):
                return f"{'D':>8} {x.name}/"

            return f"{si_unit(s.st_size):>7}B {x.name}"

    """).lstrip()
    code_block += wf.get_fn_code(
        snakefile_path_in_container(snakefile, pkg_root),
        image_name,
        wf.remote_output_url,
    )

    entrypoint = pkg_root / ".latch" / "snakemake_jit_entrypoint.py"
    entrypoint.parent.mkdir(parents=True, exist_ok=True)
    entrypoint.write_text(code_block + "\n")

    return entrypoint



import re
from dataclasses import dataclass
from enum import Enum


class Units(int, Enum):
    KiB = 2**10
    kB = 10**3

    MiB = 2**20
    MB = 10**6

    GiB = 2**30
    GB = 10**9

    TiB = 2**40
    TB = 10**12


units = Units


@dataclass(frozen=True)
class LatchConstants:
    base_image: str = (
        "812206152185.dkr.ecr.us-west-2.amazonaws.com/latch-base:fe0b-main"
    )

    file_max_size: int = 4 * units.MiB
    file_chunk_size: int = 256 * units.MiB

    # https://docs.aws.amazon.com/AmazonS3/latest/userguide/qfacts.html
    maximum_upload_parts = 10000
    maximum_upload_size = 5 * units.TiB

    pkg_name: str = "latch"
    pkg_config: str = ".latch/config"

    # todo(aidan): make this aware of the current working directory so that we do not remove useful context
    ignore_regex = re.compile(
        r"(\.git|\.latch_report\.tar\.gz|traceback\.txt|metadata\.json)$"
    )

    jump_host: str = "jump.centromere.latch.bio"
    jump_user: str = "jumpuser"

    # seconds
    centromere_poll_timeout: int = 18000
    centromere_keepalive_interval: int = 30


latch_constants = LatchConstants()


@dataclass(frozen=True)
class OAuth2Constants:
    client_id: str = "jzFBOhIbfp4EPRYZ8wmx4YyvL27LFDeB"
    """Identifies the authentication server in 0Auth2.0 flow"""

    authz_server_host: str = "https://latchai.us.auth0.com"
    """Host of the authentication server used in 0Auth2.0 flow."""

    redirect_url: str = "http://127.0.0.1:5050/callback"
    """Redirect URL registered with authentication server."""


oauth2_constants = OAuth2Constants()

docker_image_name_illegal_pat = re.compile(r"[^a-z0-9]+")



"""Latch tasks are decorators to turn functions into workflow 'nodes'.

Each task is containerized, versioned and registered with `Flyte`_ when a
workflow is uploaded to Latch. Containerized tasks are then executed on
arbitrary instances as `Kubernetes Pods`_, scheduled using `flytepropeller`_.

The type of instance that the task executes on (eg. number of available
resources, presence of GPU) can be controlled by invoking one of the set of
exported decorators.


..
    from latch import medium_task

    @medium_task
    def my_task(a: int) -> str:
        ...

.. _Kubernetes Pods:
    https://kubernetes.io/docs/concepts/workloads/pods/
.. _flytepropeller:
    https://github.com/flyteorg/flytepropeller
.. _Flyte:
    https://docs.flyte.org/en/latest/
"""

import datetime
import functools
from typing import Union
from warnings import warn

from flytekit import task
from flytekitplugins.pod import Pod
from kubernetes.client.models import (
    V1Container,
    V1PodSpec,
    V1ResourceRequirements,
    V1Toleration,
)


def _get_large_gpu_pod() -> Pod:
    """g5.8xlarge,g5.16xlarge on-demand"""

    primary_container = V1Container(name="primary")
    resources = V1ResourceRequirements(
        requests={
            "cpu": "31",
            "memory": "120Gi",
            "nvidia.com/gpu": "1",
            "ephemeral-storage": "1500Gi",
        },
        limits={
            "cpu": "64",
            "memory": "256Gi",
            "nvidia.com/gpu": "1",
            "ephemeral-storage": "2000Gi",
        },
    )
    primary_container.resources = resources

    return Pod(
        pod_spec=V1PodSpec(
            containers=[primary_container],
            tolerations=[V1Toleration(effect="NoSchedule", key="ng", value="gpu-big")],
        ),
        primary_container_name="primary",
    )


def _get_small_gpu_pod() -> Pod:
    """g4dn.2xlarge on-demand"""

    primary_container = V1Container(name="primary")
    resources = V1ResourceRequirements(
        requests={
            "cpu": "7",
            "memory": "30Gi",
            "nvidia.com/gpu": "1",
            "ephemeral-storage": "1500Gi",
        },
        limits={
            "cpu": "7",
            "memory": "30Gi",
            "nvidia.com/gpu": "1",
            "ephemeral-storage": "1500Gi",
        },
    )
    primary_container.resources = resources

    return Pod(
        pod_spec=V1PodSpec(
            containers=[primary_container],
            tolerations=[
                V1Toleration(effect="NoSchedule", key="ng", value="gpu-small")
            ],
        ),
        primary_container_name="primary",
    )


def _get_large_pod() -> Pod:
    """[ "c6i.24xlarge", "c5.24xlarge", "c5.metal", "c5d.24xlarge", "c5d.metal" ]"""

    primary_container = V1Container(name="primary")
    resources = V1ResourceRequirements(
        requests={"cpu": "90", "memory": "170Gi", "ephemeral-storage": "4500Gi"},
        limits={"cpu": "90", "memory": "170Gi", "ephemeral-storage": "4500Gi"},
    )
    primary_container.resources = resources

    return Pod(
        annotations={
            "io.kubernetes.cri-o.userns-mode": (
                "private:uidmapping=0:1048576:65536;gidmapping=0:1048576:65536"
            )
        },
        pod_spec=V1PodSpec(
            runtime_class_name="sysbox-runc",
            containers=[primary_container],
            tolerations=[
                V1Toleration(effect="NoSchedule", key="ng", value="cpu-96-spot")
            ],
        ),
        primary_container_name="primary",
    )


def _get_medium_pod() -> Pod:
    """[ "m5.8xlarge", "m5ad.8xlarge", "m5d.8xlarge", "m5n.8xlarge", "m5dn.8xlarge", "m5a.8xlarge" ]"""

    primary_container = V1Container(name="primary")
    resources = V1ResourceRequirements(
        requests={"cpu": "30", "memory": "100Gi", "ephemeral-storage": "1500Gi"},
        limits={"cpu": "30", "memory": "100Gi", "ephemeral-storage": "1500Gi"},
    )
    primary_container.resources = resources

    return Pod(
        annotations={
            "io.kubernetes.cri-o.userns-mode": (
                "private:uidmapping=0:1048576:65536;gidmapping=0:1048576:65536"
            )
        },
        pod_spec=V1PodSpec(
            runtime_class_name="sysbox-runc",
            containers=[primary_container],
            tolerations=[
                V1Toleration(effect="NoSchedule", key="ng", value="cpu-32-spot")
            ],
        ),
        primary_container_name="primary",
    )


def _get_small_pod() -> Pod:
    """any available instance"""

    primary_container = V1Container(name="primary")
    resources = V1ResourceRequirements(
        requests={"cpu": "2", "memory": "4Gi", "ephemeral-storage": "100Gi"},
        limits={"cpu": "2", "memory": "4Gi", "ephemeral-storage": "100Gi"},
    )
    primary_container.resources = resources

    return Pod(
        annotations={
            "io.kubernetes.cri-o.userns-mode": (
                "private:uidmapping=0:1048576:65536;gidmapping=0:1048576:65536"
            )
        },
        pod_spec=V1PodSpec(
            runtime_class_name="sysbox-runc",
            containers=[primary_container],
        ),
        primary_container_name="primary",
    )


large_gpu_task = functools.partial(task, task_config=_get_large_gpu_pod())
"""This task will get scheduled on a large GPU-enabled node.

This node is not necessarily dedicated to the task, but the node itself will be
on-demand.

.. list-table:: Title
   :widths: 20 20 20 20 20
   :header-rows: 1

   * - Type
     - CPU
     - RAM
     - GPU
     - On-Demand
   * - Request
     - 31
     - 120Gi
     - 1
     - True
   * - Limit
     - 64
     - 256Gi
     - 1
     - True
"""


small_gpu_task = functools.partial(task, task_config=_get_small_gpu_pod())
"""This task will get scheduled on a small GPU-enabled node.

This node will be dedicated to the task. No other tasks will be allowed to run
on it.

.. list-table:: Title
   :widths: 20 20 20 20 20
   :header-rows: 1

   * - Type
     - CPU
     - RAM
     - GPU
     - On-Demand
   * - Request
     - 7
     - 30Gi
     - 1
     - True
   * - Limit
     - 8
     - 32Gi
     - 1
     - True
"""

large_task = functools.partial(task, task_config=_get_large_pod())
"""This task will get scheduled on a large node.

This node will be dedicated to the task. No other tasks will be allowed to run
on it.

.. list-table:: Title
   :widths: 20 20 20 20 20
   :header-rows: 1

   * - Type
     - CPU
     - RAM
     - GPU
     - On-Demand
   * - Request
     - 90
     - 176Gi
     - 0
     - True
   * - Limit
     - 96
     - 196Gi
     - 0
     - True
"""


medium_task = functools.partial(task, task_config=_get_medium_pod())
"""This task will get scheduled on a medium node.

This node will be dedicated to the task. No other tasks will be allowed to run
on it.

.. list-table:: Title
   :widths: 20 20 20 20 20
   :header-rows: 1

   * - Type
     - CPU
     - RAM
     - GPU
     - On-Demand
   * - Request
     - 8
     - 32Gi
     - 0
     - True
   * - Limit
     - 12
     - 64Gi
     - 0
     - True
"""


small_task = functools.partial(task, task_config=_get_small_pod())
"""This task will get scheduled on a small node.

.. list-table:: Title
   :widths: 20 20 20 20 20
   :header-rows: 1

   * - Type
     - CPU
     - RAM
     - GPU
     - On-Demand
   * - Request
     - 2
     - 4Gi
     - 0
     - False
   * - Limit
     - 4
     - 8Gi
     - 0
     - False
"""


def custom_memory_optimized_task(cpu: int, memory: int):
    """Returns a custom task configuration requesting
    the specified CPU/RAM allocations. This task
    can utilize fewer cpu cores (62) than `custom_task`s (95)
    but can use more RAM (up to 490 GiB) than `custom_task`s (up to 179 GiB).
    This is ideal for processes which utilize a lot of memory per thread.
    Args:
        cpu: An integer number of cores to request, up to 63 cores
        memory: An integer number of Gibibytes of RAM to request, up to 511 GiB
    """
    warn(
        "`custom_memory_optimized_task` is deprecated and will be removed in a"
        " future release: use `custom_task` instead",
        DeprecationWarning,
        stacklevel=2,
    )
    if cpu > 62:
        raise ValueError(
            f"custom memory optimized task requires too many CPU cores: {cpu} (max 62)"
        )
    elif memory > 490:
        raise ValueError(
            f"custom memory optimized task requires too much RAM: {memory} GiB (max 490"
            " GiB)"
        )

    primary_container = V1Container(name="primary")
    resources = V1ResourceRequirements(
        requests={"cpu": str(cpu), "memory": f"{memory}Gi"},
        limits={"cpu": str(cpu), "memory": f"{memory}Gi"},
    )
    primary_container.resources = resources
    task_config = Pod(
        annotations={
            "io.kubernetes.cri-o.userns-mode": (
                "private:uidmapping=0:1048576:65536;gidmapping=0:1048576:65536"
            )
        },
        pod_spec=V1PodSpec(
            runtime_class_name="sysbox-runc",
            containers=[primary_container],
            tolerations=[
                V1Toleration(effect="NoSchedule", key="ng", value="mem-512-spot")
            ],
        ),
        primary_container_name="primary",
    )
    return functools.partial(task, task_config=task_config)


def custom_task(
    cpu: int,
    memory: int,
    *,
    storage_gib: int = 500,
    timeout: Union[datetime.timedelta, int] = 0,
):
    """Returns a custom task configuration requesting
    the specified CPU/RAM allocations

    Args:
        cpu: An integer number of cores to request, up to 95 cores
        memory: An integer number of Gibibytes of RAM to request, up to 490 GiB
        storage: An integer number of Gibibytes of storage to request, up to 4949 GiB
    """
    primary_container = V1Container(name="primary")
    resources = V1ResourceRequirements(
        requests={
            "cpu": str(cpu),
            "memory": f"{memory}Gi",
            "ephemeral-storage": f"{storage_gib}Gi",
        },
        limits={
            "cpu": str(cpu),
            "memory": f"{memory}Gi",
            "ephemeral-storage": f"{storage_gib}Gi",
        },
    )
    primary_container.resources = resources
    if cpu <= 31 and memory <= 127 and storage_gib <= 1949:
        task_config = Pod(
            annotations={
                "io.kubernetes.cri-o.userns-mode": (
                    "private:uidmapping=0:1048576:65536;gidmapping=0:1048576:65536"
                )
            },
            pod_spec=V1PodSpec(
                runtime_class_name="sysbox-runc",
                containers=[primary_container],
                tolerations=[
                    V1Toleration(effect="NoSchedule", key="ng", value="cpu-32-spot")
                ],
            ),
            primary_container_name="primary",
        )
    elif cpu <= 95 and memory <= 179 and storage_gib <= 4949:
        task_config = Pod(
            annotations={
                "io.kubernetes.cri-o.userns-mode": (
                    "private:uidmapping=0:1048576:65536;gidmapping=0:1048576:65536"
                )
            },
            pod_spec=V1PodSpec(
                runtime_class_name="sysbox-runc",
                containers=[primary_container],
                tolerations=[
                    V1Toleration(effect="NoSchedule", key="ng", value="cpu-96-spot")
                ],
            ),
            primary_container_name="primary",
        )
    elif cpu <= 62 and memory <= 490 and storage_gib <= 4949:
        task_config = Pod(
            annotations={
                "io.kubernetes.cri-o.userns-mode": (
                    "private:uidmapping=0:1048576:65536;gidmapping=0:1048576:65536"
                )
            },
            pod_spec=V1PodSpec(
                runtime_class_name="sysbox-runc",
                containers=[primary_container],
                tolerations=[
                    V1Toleration(effect="NoSchedule", key="ng", value="mem-512-spot")
                ],
            ),
            primary_container_name="primary",
        )
    else:
        if memory > 490:
            raise ValueError(
                f"custom task requires too much RAM: {memory} GiB (max 490 GiB)"
            )
        elif storage_gib > 4949:
            raise ValueError(
                f"custom task requires too much storage: {storage_gib} GiB (max 4949"
                " GiB)"
            )
        elif cpu > 95:
            raise ValueError(f"custom task requires too many CPU cores: {cpu} (max 95)")
        elif memory > 179 and cpu > 62:
            raise ValueError(
                f"could not resolve cpu for high memory machine: requested {cpu} cores"
                " (max 62)"
            )
        elif cpu > 62 and memory > 179:
            raise ValueError(
                f"could not resolve memory for high cpu machine: requested {memory} GiB"
                " (max 179 GiB)"
            )
        else:
            raise ValueError(
                f"custom task resource limit is too high: {cpu} (max 95) cpu cores,"
                f" {memory} GiB (max 179 GiB) memory, or {storage_gib} GiB storage (max"
                " 4949 GiB)"
            )

    return functools.partial(task, task_config=task_config, timeout=timeout)



"""
A map task lets you run a pod task or a regular task over a 
list of inputs within a single workflow node. This means you 
can run thousands of instances of the task without creating 
a node for every instance, providing valuable performance 
gains!

Some use cases of map tasks include:
 - Several inputs must run through the same code logic
 - Multiple data batches need to be processed in parallel
 - Hyperparameter optimization

Args:
    task_function: The task to be mapped, to be shown in Latch Console

Returns:
    A conditional section

Intended Use: ::

    @task
    def a_mappable_task(a: int) -> str:
        inc = a + 2
        stringified = str(inc)
        return stringified

    @task
    def coalesce(b: typing.List[str]) -> str:
        coalesced = "".join(b)
        return coalesced

    @workflow
    def my_map_workflow(a: typing.List[int]) -> str:
        mapped_out = map_task(a_mappable_task)(a=a).with_overrides(
            requests=Resources(mem="300Mi"),
            limits=Resources(mem="500Mi"),
            retries=1,
        )
        coalesced = coalesce(b=mapped_out)
        return coalesced
"""

from flytekit.core.map_task import map_task



import time
from concurrent.futures import ProcessPoolExecutor
from contextlib import closing
from dataclasses import dataclass
from itertools import repeat
from pathlib import Path
from typing import Dict, List, Set, TypedDict

import click
from latch_sdk_config.latch import config as latch_config

from latch.ldata.type import LatchPathError, LDataNodeType
from latch_cli import tinyrequests
from latch_cli.constants import Units
from latch_cli.utils import get_auth_header, human_readable_time, with_si_suffix
from latch_cli.utils.path import normalize_path

from .manager import TransferStateManager
from .node import get_node_data
from .progress import Progress, ProgressBars, get_free_index
from .utils import get_max_workers


class GetSignedUrlData(TypedDict):
    url: str


class GetSignedUrlsRecursiveData(TypedDict):
    urls: Dict[str, str]


@dataclass(frozen=True, unsafe_hash=True)
class DownloadJob:
    signed_url: str
    dest: Path


@dataclass(frozen=True)
class DownloadResult:
    num_files: int
    total_bytes: int
    total_time: int


def download(
    src: str,
    dest: Path,
    progress: Progress,
    verbose: bool,
    confirm_overwrite: bool = True,
) -> DownloadResult:
    if not dest.parent.exists():
        raise ValueError(
            f"invalid copy destination {dest}. Parent directory {dest.parent} does"
            " not exist."
        )

    normalized = normalize_path(src)
    data = get_node_data(src)
    assert src in data.data
    node_data = data.data[src]

    can_have_children = node_data.type in {
        LDataNodeType.account_root,
        LDataNodeType.dir,
        LDataNodeType.mount,
    }

    if can_have_children:
        endpoint = latch_config.api.data.get_signed_urls_recursive
    else:
        endpoint = latch_config.api.data.get_signed_url

    res = tinyrequests.post(
        endpoint,
        headers={"Authorization": get_auth_header()},
        json={"path": normalized},
    )

    if res.status_code != 200:
        err = res.json()["error"]
        msg = f"failed to fetch presigned url(s) for path {src}"
        if res.status_code == 400:
            raise ValueError(f"{msg}: download request invalid: {err}")
        if res.status_code == 401:
            raise RuntimeError(f"authorization token invalid: {err}")
        raise RuntimeError(f"{msg} with code {res.status_code}: {res.json()['error']}")

    json_data = res.json()
    if can_have_children:
        dir_data: GetSignedUrlsRecursiveData = json_data["data"]

        if dest.exists() and not normalized.endswith("/"):
            dest = dest / node_data.name

        try:
            dest.mkdir(exist_ok=True)
        except FileNotFoundError:
            raise ValueError(f"No such download destination {dest}")
        except (FileExistsError, NotADirectoryError):
            raise ValueError(f"Download destination {dest} is not a directory")

        unconfirmed_jobs: List[DownloadJob] = []
        confirmed_jobs: List[DownloadJob] = []
        rejected_jobs: Set[Path] = set()

        for rel_path, url in dir_data["urls"].items():
            unconfirmed_jobs.append(DownloadJob(url, dest / rel_path))

        for job in unconfirmed_jobs:
            reject_job = False
            for parent in job.dest.parents:
                if parent in rejected_jobs:
                    reject_job = True
                    break

            if reject_job:
                continue

            try:
                job.dest.parent.mkdir(parents=True, exist_ok=True)
                confirmed_jobs.append(job)
            except FileExistsError:
                if confirm_overwrite and click.confirm(
                    f"A file already exists at {job.dest.parent}. Overwrite?",
                    default=False,
                ):
                    job.dest.parent.unlink()
                    job.dest.parent.mkdir(parents=True, exist_ok=True)
                    confirmed_jobs.append(job)
                else:
                    print(f"Skipping {job.dest.parent}, file already exists")
                    rejected_jobs.add(job.dest.parent)

        num_files = len(confirmed_jobs)

        if progress == Progress.none:
            num_bars = 0
            show_total_progress = False
        elif progress == Progress.total:
            num_bars = 0
            show_total_progress = True
        else:
            num_bars = min(get_max_workers(), num_files)
            show_total_progress = True

        with TransferStateManager() as manager:
            progress_bars: ProgressBars
            with closing(
                manager.ProgressBars(
                    num_bars,
                    show_total_progress=show_total_progress,
                    verbose=verbose,
                )
            ) as progress_bars:
                progress_bars.set_total(num_files, "Copying Files")

                start = time.monotonic()

                # todo(ayush): benchmark this against asyncio
                with ProcessPoolExecutor(max_workers=get_max_workers()) as executor:
                    total_bytes = sum(
                        executor.map(
                            download_file,
                            confirmed_jobs,
                            repeat(progress_bars),
                        )
                    )

                end = time.monotonic()
    else:
        file_data: GetSignedUrlData = json_data["data"]

        num_files = 1

        if dest.exists() and dest.is_dir():
            dest = dest / node_data.name

        if progress == Progress.none:
            num_bars = 0
        else:
            num_bars = 1

        with TransferStateManager() as manager:
            progress_bars: ProgressBars
            with closing(
                manager.ProgressBars(
                    num_bars,
                    show_total_progress=False,
                    verbose=verbose,
                )
            ) as progress_bars:
                start = time.monotonic()
                total_bytes = download_file(
                    DownloadJob(file_data["url"], dest),
                    progress_bars,
                )
                end = time.monotonic()

    total_time = end - start

    return DownloadResult(num_files, total_bytes, total_time)


# dest will always be a path which includes the copied file as its leaf
# e.g. download_file("a/b.txt", Path("c/d.txt")) will copy the content of 'b.txt' into 'd.txt'
def download_file(
    job: DownloadJob,
    progress_bars: ProgressBars,
) -> int:
    # todo(ayush): benchmark parallelized downloads using the range header
    with open(job.dest, "wb") as f:
        res = tinyrequests.get(job.signed_url, stream=True)
        if res.status_code != 200:
            raise RuntimeError(
                f"failed to download {job.dest.name}: {res.status_code}:"
                f" {res.json()['error']}"
            )

        total_bytes = res.headers.get("Content-Length")
        assert total_bytes is not None, "Must have a content-length header"

        with get_free_index(progress_bars) as pbar_index:
            progress_bars.set(
                index=pbar_index, total=int(total_bytes), desc=job.dest.name
            )

            start = time.monotonic()
            try:
                for data in res.iter_content(
                    chunk_size=5 * Units.MiB
                ):  # todo(ayush): figure out why chunk_size = None breaks in pods
                    f.write(data)
                    progress_bars.update(pbar_index, len(data))
            finally:
                end = time.monotonic()
                progress_bars.update_total_progress(1)
                progress_bars.write(
                    f"Downloaded {job.dest.name} ({with_si_suffix(int(total_bytes))})"
                    f" in {human_readable_time(end - start)}"
                )

        return int(total_bytes)



import re
import sys
import traceback
from dataclasses import dataclass
from pathlib import Path
from textwrap import dedent
from typing import Dict, Optional, Tuple

import click
import docker
import paramiko
import paramiko.util
from docker.transport import SSHHTTPAdapter
from flytekit.core.base_task import PythonTask
from flytekit.core.context_manager import FlyteEntities
from flytekit.core.workflow import PythonFunctionWorkflow
from latch_sdk_config.latch import config

import latch_cli.tinyrequests as tinyrequests
from latch_cli.centromere.utils import (
    RemoteConnInfo,
    _construct_dkr_client,
    _construct_ssh_client,
    _import_flyte_objects,
)
from latch_cli.constants import docker_image_name_illegal_pat
from latch_cli.docker_utils import get_default_dockerfile
from latch_cli.utils import (
    WorkflowType,
    account_id_from_token,
    current_workspace,
    generate_temporary_ssh_credentials,
    hash_directory,
    retrieve_or_login,
)


@dataclass
class _Container:
    dockerfile: Path
    pkg_dir: Path
    image_name: str


class _CentromereCtx:
    """Manages state for interaction with centromere.

    The context holds values that are relevant throughout the "lifetime" of a
    registration or remote execution, eg. location of local code and
    package name, as well as managing docker, ssh clients.
    """

    dkr_repo: Optional[str] = None
    dkr_client: Optional[docker.APIClient] = None
    ssh_client: Optional[paramiko.SSHClient] = None
    pkg_root: Optional[Path] = None  # root
    disable_auto_version: bool = False
    image_full = None
    version = None
    serialize_dir = None
    default_container: _Container
    workflow_type: WorkflowType
    snakefile: Optional[Path]

    latch_register_api_url = config.api.workflow.register
    latch_image_api_url = config.api.workflow.upload_image
    latch_provision_url = config.api.centromere.provision
    latch_get_image_url = config.api.workflow.get_image
    latch_check_version_url = config.api.workflow.check_version

    ssh_key_path: Optional[Path] = None
    jump_key_path: Optional[Path] = None
    ssh_config_path: Optional[Path] = None

    internal_ip: Optional[str] = None
    username: Optional[str] = None

    def __init__(
        self,
        pkg_root: Path,
        *,
        disable_auto_version: bool = False,
        remote: bool = False,
        snakefile: Optional[Path] = None,
        use_new_centromere: bool = False,
    ):
        self.use_new_centromere = use_new_centromere
        self.remote = remote
        self.disable_auto_version = disable_auto_version

        try:
            self.token = retrieve_or_login()
            self.account_id = current_workspace()

            self.dkr_repo = config.dkr_repo
            self.pkg_root = pkg_root.resolve()

            if snakefile is None:
                self.workflow_type = WorkflowType.latchbiosdk
            else:
                self.workflow_type = WorkflowType.snakemake
                self.snakefile = snakefile

            self.container_map: Dict[str, _Container] = {}
            if self.workflow_type == WorkflowType.latchbiosdk:
                _import_flyte_objects([self.pkg_root])
                for entity in FlyteEntities.entities:
                    if isinstance(entity, PythonFunctionWorkflow):
                        self.workflow_name = entity.name

                    if isinstance(entity, PythonTask):
                        if (
                            hasattr(entity, "dockerfile_path")
                            and entity.dockerfile_path is not None
                        ):
                            self.container_map[entity.name] = _Container(
                                dockerfile=entity.dockerfile_path,
                                image_name=self.task_image_name(entity.name),
                                pkg_dir=entity.dockerfile_path.parent,
                            )
                if not hasattr(self, "workflow_name"):
                    click.secho(
                        dedent("""
                                     Unable to locate workflow code. If you
                                     are a registering a Snakemake project,
                                     make sure to pass the Snakefile path with
                                     the --snakefile flag."""),
                        fg="red",
                    )
                    raise click.exceptions.Exit(1)
            else:
                assert snakefile is not None

                import latch.types.metadata as metadata

                from ..services.register.utils import import_module_by_path
                from ..snakemake.serialize import (
                    get_snakemake_metadata_example,
                    snakemake_workflow_extractor,
                )
                from ..snakemake.utils import load_snakemake_metadata

                meta_file = load_snakemake_metadata(pkg_root)
                if meta_file is not None:
                    click.echo(
                        f"Using metadata file {click.style(meta_file, italic=True)}"
                    )
                else:
                    new_meta = pkg_root / "latch_metadata" / "__init__.py"
                    click.echo("Trying to extract metadata from the Snakefile")
                    try:
                        snakemake_workflow_extractor(pkg_root, snakefile)
                    except (ImportError, FileNotFoundError):
                        traceback.print_exc()
                        click.secho(
                            "\n\n\n"
                            + "The above error occured when reading "
                            + "the Snakefile to extract workflow metadata.",
                            bold=True,
                            fg="red",
                        )
                        click.secho(
                            "\nIt is possible to avoid including the Snakefile"
                            " prior to registration by providing a"
                            " `latch_metadata.py` file in the workflow root.\nThis"
                            " way it is not necessary to install dependencies or"
                            " ensure that Snakemake inputs locally.",
                            fg="red",
                        )
                        click.secho("\nExample ", fg="red", nl=False)

                        snakemake_metadata_example = get_snakemake_metadata_example(
                            pkg_root.name
                        )
                        click.secho(f"`{new_meta}`", bold=True, fg="red", nl=False)
                        click.secho(
                            f" file:\n```\n{snakemake_metadata_example}```",
                            fg="red",
                        )
                        if click.confirm(
                            click.style(
                                "Generate example metadata file now?",
                                bold=True,
                                fg="red",
                            ),
                            default=True,
                        ):
                            new_meta.write_text(snakemake_metadata_example)

                            import platform

                            system = platform.system()
                            if system in {
                                "Windows",
                                "Linux",
                                "Darwin",
                            } and click.confirm(
                                click.style(
                                    "Open the generated file?", bold=True, fg="red"
                                ),
                                default=True,
                            ):
                                import subprocess

                                if system == "Linux":
                                    res = subprocess.run(
                                        ["xdg-open", new_meta]
                                    ).returncode
                                elif system == "Darwin":
                                    res = subprocess.run(["open", new_meta]).returncode
                                elif system == "Windows":
                                    import os

                                    res = os.system(str(new_meta.resolve()))
                                else:
                                    res = None

                                if res is not None and res != 0:
                                    click.secho("Failed to open file", fg="red")
                        sys.exit(1)

                if metadata._snakemake_metadata is None:
                    click.secho(
                        dedent(
                            """
                        Make sure a `latch_metadata.py` file exists in the Snakemake
                         project root.""",
                        ),
                        fg="red",
                    )
                    raise click.exceptions.Exit(1)

                # todo(kenny): support per container task and custom workflow
                # name for snakemake
                self.workflow_name = f"{metadata._snakemake_metadata.name}_jit_register"

            version_file = self.pkg_root / "version"
            try:
                self.version = version_file.read_text()
            except FileNotFoundError:
                self.version = "0.1.0"
                version_file.write_text(f"{self.version}\n")
                click.echo(
                    f"Created a version file with initial version {self.version}."
                )
            self.version = self.version.strip()

            if not self.disable_auto_version:
                hash = hash_directory(self.pkg_root)
                self.version = f"{self.version}-{hash[:6]}"
                click.echo(f"  {self.version}\n")

            if self.nucleus_check_version(self.version, self.workflow_name):
                click.secho(
                    f"\nVersion ({self.version}) already exists."
                    " Make sure that you've saved any changes you made.",
                    fg="red",
                    bold=True,
                )
                sys.exit(1)

            self.default_container = _Container(
                dockerfile=get_default_dockerfile(
                    self.pkg_root, wf_type=self.workflow_type
                ),
                image_name=self.image_tagged,
                pkg_dir=self.pkg_root,
            )

            if remote:
                # todo(maximsmol): connect only AFTER confirming registration
                self.ssh_key_path = self.pkg_root / ".latch/ssh_key"
                self.jump_key_path = self.pkg_root / ".latch/jump_key"
                self.public_key = generate_temporary_ssh_credentials(
                    self.ssh_key_path, add_to_agent=False
                )

                if use_new_centromere:
                    self.internal_ip, self.username = (
                        self.provision_register_deployment()
                    )
                else:
                    self.internal_ip, self.username = self.get_old_centromere_info()

                remote_conn_info = RemoteConnInfo(
                    ip=self.internal_ip,
                    username=self.username,
                    jump_key_path=self.jump_key_path,
                    ssh_key_path=self.ssh_key_path,
                )

                ssh_client = _construct_ssh_client(
                    remote_conn_info, use_gateway=use_new_centromere
                )
                self.ssh_client = ssh_client

                def _patched_connect(self): ...

                def _patched_create_paramiko_client(self, base_url):
                    self.ssh_client = ssh_client

                SSHHTTPAdapter._create_paramiko_client = _patched_create_paramiko_client
                SSHHTTPAdapter._connect = _patched_connect

                self.dkr_client = _construct_dkr_client(ssh_host="ssh://fake")

            else:
                self.dkr_client = _construct_dkr_client()
        except (Exception, KeyboardInterrupt) as e:
            self.cleanup()
            raise e

    @property
    def image(self):
        """The image to be registered."""
        if self.account_id is None:
            raise ValueError("You need to log in before you can register a workflow.")

        # CAUTION ~ this weird formatting is maintained indepedently in the
        # nucleus endpoint and here.
        # Name for federated token request has minimum of 2 characters.
        if int(self.account_id) < 10:
            account_id = f"x{self.account_id}"
        else:
            account_id = self.account_id

        from ..utils import identifier_suffix_from_str

        wf_name = identifier_suffix_from_str(self.workflow_name).lower()
        wf_name = docker_image_name_illegal_pat.sub("_", wf_name)

        return f"{account_id}_{wf_name}"

    @property
    def image_tagged(self):
        """The tagged image to be registered.

        eg. dkr.ecr.us-west-2.amazonaws.com/pkg_name:version
        """
        if self.version is None:
            raise ValueError(
                "Attempting to create a tagged image name without first "
                "extracting the package version."
            )

        # From AWS:
        #   A tag name must be valid ASCII and may contain lowercase and uppercase letters,
        #   digits, underscores, periods and dashes. A tag name may not start with a period
        #   or a dash and may contain a maximum of 128 characters.

        match = re.match("^[a-zA-Z0-9_][a-zA-Z0-9._-]{,127}$", self.version)
        if match is None:
            raise ValueError(
                f"{self.version} is an invalid version for AWS "
                "ECR. Please provide a version that accomodates the "
                "tag restrictions listed here - "
                "https://docs.aws.amazon.com/AmazonECR/latest/userguide/ecr-using-tags.html"
            )

        if self.image is None or self.version is None:
            raise ValueError(
                "Attempting to create a tagged image name without first "
                " logging in or extracting the package version."
            )
        return f"{self.image}:{self.version}"

    def task_image_name(self, task_name: str) -> str:
        return f"{self.image}:{task_name}-{self.version}"

    @property
    def full_image(self):
        """The full image to be registered (without a tag).

            <repo/image>


        An example: ::

            dkr.ecr.us-west-2.amazonaws.com/pkg_name

        """
        return f"{self.dkr_repo}/{self.image}"

    def get_old_centromere_info(self) -> Tuple[str, str]:
        headers = {"Authorization": f"Bearer {self.token}"}

        response = tinyrequests.post(
            self.latch_provision_url,
            headers=headers,
            json={
                "public_key": self.public_key,
            },
        )

        resp = response.json()
        try:
            public_ip = resp["ip"]
            username = resp["username"]
        except KeyError as e:
            raise ValueError(
                f"Malformed response from request for access token {resp}"
            ) from e

        return public_ip, username

    def provision_register_deployment(self) -> Tuple[str, str]:
        """Retrieve centromere IP + username."""
        click.echo("Provisioning register instance. This may take a few minutes.")

        assert self.ssh_key_path is not None
        assert self.jump_key_path is not None

        resp = tinyrequests.post(
            "https://centromere.latch.bio/register/start",
            headers={"Authorization": f"Latch-SDK-Token {self.token}"},
            json={"SSHKey": self.ssh_key_path.with_suffix(".pub").read_text()},
        )

        json_data = resp.json()
        if resp.status_code != 200:
            raise ValueError(json_data["Error"])

        hostname = json_data["InternalHost"]
        self.jump_key_path.write_text(json_data["JumpKey"])
        self.jump_key_path.chmod(0o600)

        self.centromere_hostname = hostname

        return hostname, "root"

    def downscale_register_deployment(self):
        if not (self.remote and self.use_new_centromere):
            return

        resp = tinyrequests.post(
            "https://centromere.latch.bio/register/stop",
            headers={"Authorization": f"Latch-SDK-Token {self.token}"},
            json={"InternalHostName": self.centromere_hostname},
        )

        if resp.status_code != 200:
            raise ValueError("unable to downscale register deployment")

    def nucleus_get_image(self, task_name: str, version: Optional[str] = None) -> str:
        """Retrieve fqn of the container for a task and optional version."""

        headers = {"Authorization": f"Bearer {self.token}"}
        response = tinyrequests.post(
            self.latch_get_image_url,
            headers=headers,
            json={
                "task_name": task_name,
            },
        )

        resp = response.json()
        try:
            return resp["image_name"]
        except KeyError as e:
            raise ValueError(
                f"Malformed response from request for access token {resp}"
            ) from e

    def nucleus_check_version(self, version: str, workflow_name: str) -> bool:
        """Check if version has already been registered for given workflow"""

        headers = {"Authorization": f"Bearer {self.token}"}

        ws_id = current_workspace()
        if ws_id is None or ws_id == "":
            ws_id = account_id_from_token(retrieve_or_login())

        response = tinyrequests.post(
            self.latch_check_version_url,
            headers=headers,
            json={
                "version": version,
                "workflow_name": workflow_name,
                "ws_account_id": ws_id,
            },
        )

        resp = response.json()
        try:
            return resp["exists"]
        except KeyError as e:
            raise ValueError(
                f"Malformed response from request for access token {resp}"
            ) from e

    def __enter__(self):
        return self

    def cleanup(self):
        if self.ssh_key_path is not None:
            self.ssh_key_path.unlink(missing_ok=True)
            self.ssh_key_path.with_suffix(".pub").unlink(missing_ok=True)

        if self.jump_key_path is not None:
            self.jump_key_path.unlink(missing_ok=True)

        self.downscale_register_deployment()

    def __exit__(self, type, value, traceback):
        self.cleanup()



from flytekit.core.launch_plan import reference_launch_plan

from ..types.directory import LatchDir
from ..types.file import LatchFile


@reference_launch_plan(
    project="1",
    domain="development",
    name="wf.__init__.gene_ontology_pathway_analysis",
    version="0.0.43",
)
def gene_ontology_pathway_analysis(
    contrast_csv: LatchFile,
    report_name: str,
    number_of_pathways: int = 20,
    output_location: LatchDir = LatchDir("latch:///Pathway Analysis/"),
) -> LatchDir: ...



"""Service to launch a workflow."""

import importlib.util
import typing
from pathlib import Path
from typing import Optional, Tuple, Union

import google.protobuf.json_format as gpjson
import requests
from flyteidl.core.types_pb2 import LiteralType
from flytekit.core.context_manager import FlyteContextManager
from flytekit.core.type_engine import TypeEngine
from latch_sdk_config.latch import config
from requests.adapters import HTTPAdapter
from urllib3.util.retry import Retry

from latch_cli.utils import current_workspace, retrieve_or_login


def launch(params_file: Path, version: Optional[str] = None) -> str:
    """Launches a (versioned) workflow with parameters specified in python.

    Using a parameter map written in python (this can be generated for you with
    `get_params`), this function will launch the workflow specified in the file
    using the parameters therein. This function also accepts an optional
    `version` parameter to further specify the workflow to be run. If it is not
    provided, this function will default to running the latest version of the
    specified workflow.

    Args:
        params_file: A path pointing to a python parameter file containing a
            function call that represents the workflow execution with valid
            parameter values.
        version: An optional workflow version to launch, defaulting to the
            latest if not provided.

    Returns:
        The name of the workflow.

    Example:
        >>> launch(Path("wf.__init__.assemble_and_sort.params.py"))
            # Launches an execution of `wf.__init__.assemble_and_sort` with the
            # parameters specified in the referenced file.
    """

    token = retrieve_or_login()

    with open(params_file, "r") as pf:
        param_code = pf.read()
        spec = importlib.util.spec_from_loader("wf_params", loader=None)
        param_module = importlib.util.module_from_spec(spec)
        exec(param_code, param_module.__dict__)

    module_vars = vars(param_module)
    try:
        wf_params = module_vars["params"]
    except KeyError as e:
        raise ValueError(
            f"Execution file {params_file.name} needs to have"
            " a parameter value dictionary named 'params'"
        ) from e

    wf_name = wf_params.get("_name")
    if wf_name is None:
        raise ValueError(
            f"The dictionary of parameters in the launch file lacks the"
            f" _name key used to identify the workflow. Make sure a _name"
            f" key with the workflow name exists in the dictionary."
        )

    wf_id, wf_interface, _ = _get_workflow_interface(token, wf_name, version)

    wf_vars = wf_interface["variables"]
    wf_literals = {}
    for key, value in wf_vars.items():
        ctx = FlyteContextManager.current_context()
        literal_type_json = value["type"]
        literal_type = gpjson.ParseDict(literal_type_json, LiteralType())

        if key in wf_params:
            python_value = wf_params[key]
            # Recover parameterized generics for TypeTransformer.
            python_type = _guess_python_type(python_value)

            python_type_literal = TypeEngine.to_literal(
                ctx, python_value, python_type, literal_type
            )

            wf_literals[key] = gpjson.MessageToDict(python_type_literal.to_flyte_idl())

    _launch_workflow(token, wf_id, wf_literals)
    return wf_name


def _guess_python_type(v: any) -> typing.T:
    """Python literal guesser.

    We will attempt to construct the correct python type representation from the
    value and JSON type representation and rely on the TypeTransformer to produce
    the correct flyte literal for execution (FlyteIDL representation of the value).
    This is essentially how flytekit does it.

    Using the type() function alone is not sufficient because flyte interprets
    the python list literal as a generic collection type and needs a
    parameterization.

    For example:

    ..
       >> type(["AUG", "AAA"]) = list
       <class 'list'>

    Becomes List[str] s.t.

    ..
       >> TypeEngine.to_literal(ctx, ["AUG", "AAA"], List[str], type_literal)

    Returns our desired flyte literal.
    """

    if type(v) is list:
        if len(v) == 0:
            return typing.List[None]
        elif type(v[0]) is list:
            return typing.List[_guess_python_type(v[0])]
        else:
            return typing.List[type(v[0])]

    # TODO: maps, Records, future complex types

    return type(v)


def _get_workflow_interface(
    token: str, wf_name: str, version: Union[None, str]
) -> Tuple[int, dict]:
    """Retrieves the set of idl parameter values for a given workflow by name.

    Returns workflow id + interface as JSON string.
    """

    headers = {"Authorization": f"Bearer {token}"}
    _interface_request = {
        "workflow_name": wf_name,
        "version": version,
        "ws_account_id": current_workspace(),
    }

    url = config.api.workflow.interface

    # TODO(ayush) - figure out why timeout within this endpoint only.
    session = requests.Session()
    retries = 5
    retry = Retry(
        total=retries,
        read=retries,
        connect=retries,
        method_whitelist=False,
    )
    adapter = HTTPAdapter(max_retries=retry)
    session.mount("http://", adapter)
    session.mount("https://", adapter)

    response = session.post(url, headers=headers, json=_interface_request)

    wf_interface_resp = response.json()

    wf_id, wf_interface, wf_default_params = (
        wf_interface_resp.get("id"),
        wf_interface_resp.get("interface"),
        wf_interface_resp.get("default_params"),
    )
    if wf_interface is None:
        raise ValueError(
            "Could not find interface. Nucleus returned a malformed JSON response -"
            f" {wf_interface_resp}"
        )
    if wf_id is None:
        raise ValueError(
            "Could not find wf ID. Nucleus returned a malformed JSON response -"
            f" {wf_interface_resp}"
        )
    if wf_default_params is None:
        raise ValueError(
            "Could not find wf default parameters. Nucleus returned a malformed JSON"
            f" response - {wf_interface_resp}"
        )

    return int(wf_id), wf_interface, wf_default_params


def _launch_workflow(token: str, wf_id: str, params: dict) -> bool:
    """Launch the workflow of given id with parameter map.

    Return True if success
    """

    # TODO (kenny) - pull out to consolidated requests class
    # Server sometimes stalls on requests with python user-agent
    headers = {
        "Authorization": f"Bearer {token}",
        "User-Agent": (
            "Mozilla/5.0 (Windows NT 6.1; Win64; x64) AppleWebKit/537.36 (KHTML, like"
            " Gecko) Chrome/72.0.3626.119 Safari/537.36"
        ),
    }

    _interface_request = {
        "workflow_id": str(wf_id),
        "params": params,
        "ws_account_id": current_workspace(),
    }
    url = config.api.execution.create

    response = requests.post(url, headers=headers, json=_interface_request)

    if response.status_code == 403:
        raise PermissionError(
            "You need access to the latch sdk beta ~ join the waitlist @"
            " https://latch.bio/sdk"
        )
    elif response.status_code == 401:
        raise ValueError(
            "your token has expired - please run latch login to refresh your token and"
            " try again."
        )
    wf_interface_resp = response.json()

    return wf_interface_resp.get("success") is True



"""Service to initialize boilerplate."""

import re
import shutil
import subprocess
from pathlib import Path
from typing import Callable, Optional

import click

from latch_cli.docker_utils import generate_dockerfile
from latch_cli.menus import select_tui
from latch_cli.workflow_config import BaseImageOptions, create_and_write_config


def _get_boilerplate(pkg_root: Path, source_path: Path, copy_wf_dir: bool = True):
    pkg_root = pkg_root.resolve()
    source_path = source_path.resolve()

    wf_root = pkg_root / "wf"
    wf_root.mkdir(exist_ok=True)

    if copy_wf_dir is True:
        for f in source_path.glob("*.py"):
            shutil.copy(f, wf_root)

    pkg_root_globs = [
        "LICENSE*",
        "README*",
        "*requirements*",
        "env*",
        "Dockerfile*",
    ]

    for g in pkg_root_globs:
        for f in source_path.glob(g):
            shutil.copy(f, pkg_root)

    if (source_path / ".env").exists():
        shutil.copy(source_path / ".env", pkg_root)

    common_source = source_path.parent / "common"
    for f in common_source.iterdir():
        shutil.copy(f, pkg_root)

    version_f = pkg_root / "version"
    with open(version_f, "w") as f:
        f.write("0.0.0")


def _get_example_reference(pkg_root: Path):
    import boto3
    from botocore import UNSIGNED
    from botocore.config import Config

    pkg_root = pkg_root.resolve()

    data_root = pkg_root / "reference"
    data_root.mkdir(exist_ok=True)

    ref_ids = [
        "wuhan.1.bt2",
        "wuhan.2.bt2",
        "wuhan.3.bt2",
        "wuhan.4.bt2",
        "wuhan.fasta",
        "wuhan.rev.1.bt2",
        "wuhan.rev.2.bt2",
    ]

    s3 = boto3.client("s3", config=Config(signature_version=UNSIGNED))

    print("Downloading workflow data ", flush=True, end="")
    for id in ref_ids:
        print(".", flush=True, end="")
        with open(data_root / id, "wb") as f:
            s3.download_fileobj("latch-public", f"sdk/{id}", f)
    print()


def _gen_assemble_and_sort(pkg_root: Path):
    pkg_root = pkg_root.resolve()
    source_path = Path(__file__).parent / "assemble_and_sort"

    _get_boilerplate(pkg_root, source_path)

    _get_example_reference(pkg_root)

    print("Downloading bowtie2")
    bowtie2_base_name = "bowtie2-2.5.1-linux-x86_64"
    subprocess.run(
        [
            "curl",
            f"https://latch-public.s3.us-west-2.amazonaws.com/sdk/{bowtie2_base_name}.zip",
            "-o",
            str(pkg_root / f"{bowtie2_base_name}.zip"),
        ],
        check=True,
    )
    subprocess.run(
        ["unzip", str(pkg_root / f"{bowtie2_base_name}.zip"), "-d", str(pkg_root)],
        check=True,
    )

    bowtie_dir = pkg_root / "bowtie2"
    bowtie_dir.mkdir(exist_ok=True)

    subprocess.run(
        f"mv {str(pkg_root / bowtie2_base_name)}/*bowtie2* {str(pkg_root / 'bowtie2')}",
        check=True,
        shell=True,
    )

    paths_to_remove = [
        pkg_root / bowtie2_base_name,
        pkg_root / f"{bowtie2_base_name}.zip",
    ]
    paths_to_remove.extend((pkg_root / "bowtie2").glob("*-debug*"))

    for f in paths_to_remove:
        if f.is_file():
            f.unlink()
        else:
            shutil.rmtree(str(f))


def _gen_template(pkg_root: Path):
    pkg_root = pkg_root.resolve()
    source_path = Path(__file__).parent / "template"

    _get_boilerplate(pkg_root, source_path)

    wf_metadata_params = {
        "WF_NAME": click.prompt(
            "Workflow Name", default="CHANGE ME", show_default=False
        ),
        "AUTHOR_NAME": click.prompt(
            "Author Name", default="CHANGE ME", show_default=False
        ),
    }

    init_file = Path(pkg_root / "wf" / "__init__.py")
    lines = init_file.read_text()
    init_file.unlink()

    for k, v in wf_metadata_params.items():
        lines = lines.replace(k, v)

    init_file.write_text(lines)


def _gen_example_r(pkg_root: Path):
    pkg_root = pkg_root.resolve()
    source_path = Path(__file__).parent / "example_r"

    _get_boilerplate(pkg_root, source_path)


def _gen_example_conda(pkg_root: Path):
    pkg_root = pkg_root.resolve()
    source_path = Path(__file__).parent / "example_conda"

    _get_boilerplate(pkg_root, source_path)

    conda_env_dest = pkg_root / "environment.yaml"
    conda_env_src = source_path / "environment.yaml"
    shutil.copy(conda_env_src, conda_env_dest)


def _gen_example_docker(pkg_root: Path):
    pkg_root = pkg_root.resolve()
    source_docker_path = Path(__file__).parent / "example_docker"

    _get_boilerplate(pkg_root, source_docker_path)


def _gen_example_snakemake(pkg_root: Path):
    pkg_root = pkg_root.resolve()
    source_path = Path(__file__).parent / "example_snakemake"

    _get_boilerplate(pkg_root, source_path, copy_wf_dir=False)

    snakefile_dest = pkg_root / "Snakefile"
    snakefile_src = source_path / "Snakefile"
    shutil.copy(snakefile_src, snakefile_dest)

    snakefile_dest = pkg_root / "latch_metadata.py"
    snakefile_src = source_path / "latch_metadata.py"
    shutil.copy(snakefile_src, snakefile_dest)

    print("Downloading data")
    snakemake_base_name = "snakemake_tutorial_data"
    subprocess.run(
        [
            "curl",
            f"https://latch-public.s3.us-west-2.amazonaws.com/sdk/{snakemake_base_name}.zip",
            "-o",
            str(pkg_root / f"{snakemake_base_name}.zip"),
        ],
        check=True,
    )
    subprocess.run(
        ["unzip", str(pkg_root / f"{snakemake_base_name}.zip"), "-d", str(pkg_root)],
        check=True,
    )

    scripts_dest = pkg_root / "scripts"
    scripts_src = source_path / "scripts"
    shutil.copytree(scripts_src, scripts_dest)


def _gen_example_nfcore(pkg_root: Path):
    pkg_root = pkg_root.resolve()
    source_path = Path(__file__).parent / "example_nfcore"

    _get_boilerplate(pkg_root, source_path)


option_map = {
    "Empty workflow": _gen_template,
    "Subprocess Example": _gen_assemble_and_sort,
    "R Example": _gen_example_r,
    "Conda Example": _gen_example_conda,
    "Docker Example": _gen_example_docker,
    "Snakemake Example": _gen_example_snakemake,
    "NFCore Example": _gen_example_nfcore,
}


template_flag_to_option = {
    "empty": "Empty workflow",
    "docker": "Docker Example",
    "subprocess": "Subprocess Example",
    "r": "R Example",
    "conda": "Conda Example",
    "snakemake": "Snakemake Example",
    "nfcore": "NFCore Example",
}


base_docker_image_options = {
    "Default Latch Docker image with No Dependencies": BaseImageOptions.default,
    "Latch Docker image with Nvidia CUDA/cuDNN (cuda 11.4.2, cudnn 8) drivers": (
        BaseImageOptions.cuda
    ),
    "Latch Docker image with OpenCL (ubuntu 18.04) drivers": BaseImageOptions.opencl,
}


def init(
    pkg_name: str,
    template: Optional[str],
    expose_dockerfile: bool = True,
    base_image_type_str: str = "default",
) -> bool:
    """Creates boilerplate workflow files in the user's working directory.

    Args:
        pkg_name: A identifier for the workflow - will name the boilerplate
            directory as well as functions within the constructed package.
        template: A template to use for the workflow. If None, you will be
            prompted to choose a template.
                * "empty": An empty workflow wrapper
                * "subprocess": An example workflow that runs a subprocess
                * "r": A template workflow for executing an R script
                * "conda": A template workflow for executing code within a conda environment
        expose_dockerfile: Whether to expose a Dockerfile in the workflow.
            If true, the Dockerfile will be created at init time and can be
            modified. Otherwise, the Dockerfile will be created at registration
            time and the user will not be able to modify it. At any point,
            the user can switch modes by executing `latch dockerfile .` in
            the workflow directory.
        base_image_type_str: Base image to use for the workflow. Default value
            is "default". The following options are available:
                * "default": with no additional dependencies
                * "cuda": with Nvidia CUDA/cuDNN (cuda 11.4.2, cudnn 8) drivers
                * "opencl": with OpenCL (ubuntu 18.04) drivers
                * "docker": with the Docker daemon
    Example:

        >>> init("test-workflow", "empty", False)
            # The resulting file structure will look like
            #   test-workflow
            #   ├── version
            #   └── wf
            #       ├── __init__.py
            #       └── task.py

    """

    pkg_root = Path(pkg_name).resolve()
    pkg_name = pkg_root.name

    append_ctx_to_error: Callable[[str], str] = lambda x: (
        f"{x}. Current directory name: {pkg_root}"
        if pkg_root == Path.cwd()
        else f"{x}. Supplied name: {pkg_root}"
    )

    # Workflow name must not contain capitals or start or end in a hyphen or underscore. If it does, we should throw an error.

    if any(char.isupper() for char in pkg_name):
        raise ValueError(
            append_ctx_to_error(
                f"package name must not contain any upper-case characters: {pkg_name}"
            ),
        )

    if re.search("^[a-z]", pkg_name) is None:
        raise ValueError(
            append_ctx_to_error(
                f"package name must start with a lower-case letter: {pkg_name}"
            ),
        )

    if re.search("[a-z]$", pkg_name) is None:
        raise ValueError(
            append_ctx_to_error(
                f"package name must end with a lower-case letter: {pkg_name}"
            ),
        )

    for char in pkg_name:
        if not char.isalnum and char not in ["-", "_"]:
            raise ValueError(
                append_ctx_to_error(
                    "package name must only contain alphanumeric characters, hyphens,"
                    f" and underscores: found `{char}`."
                ),
            )

    if template is None:
        template_func = select_tui(
            title="Select Workflow Template",
            options=[
                {"display_name": name, "value": fn} for name, fn in option_map.items()
            ],
        )
    else:
        template_func = option_map[template_flag_to_option[template]]

    if template_func is None:
        return False

    try:
        pkg_root.mkdir(parents=True)
    except FileExistsError:
        if not pkg_root.is_dir():
            raise ValueError(
                f"Cannot create directory `{pkg_name}`. A file with that name already"
                " exists."
            )

        if not click.confirm(
            f"Warning -- existing files in directory `{pkg_name}` may be overwritten by"
            " boilerplate. Continue?"
        ):
            return False

    base_image_type = BaseImageOptions[base_image_type_str]

    if template_func == "Empty workflow":
        base_image_type = select_tui(
            title="Select the base docker image to use for the workflow",
            options=[
                {"display_name": name, "value": value}
                for name, value in base_docker_image_options.items()
            ],
        )

        if base_image_type is None:
            return False

    template_func(pkg_root)

    create_and_write_config(pkg_root, base_image_type)

    if expose_dockerfile:
        generate_dockerfile(pkg_root, pkg_root / "Dockerfile")

    return True



from typing import Annotated, List, Optional

from flytekit.core.annotation import FlyteAnnotation
from flytekit.core.launch_plan import reference_launch_plan

from ..types.directory import LatchDir
from ..types.file import LatchFile


@reference_launch_plan(
    project="4107",
    domain="development",
    name="wf.__init__.deseq2_wf",
    version="1.3.18-7a4e16",
)
def deseq2_wf(
    report_name: str,
    count_table_source: str = "single",
    raw_count_table: Optional[
        Annotated[
            LatchFile,
            FlyteAnnotation(
                {
                    "_tmp_hack_deseq2": "counts_table",
                    "rules": [
                        {
                            "regex": r".*\.(csv|tsv|xlsx)$",
                            "message": "Expected a CSV, TSV, or XLSX file",
                        }
                    ],
                }
            ),
        ]
    ] = None,
    raw_count_tables: List[LatchFile] = [],
    count_table_gene_id_column: str = "gene_id",
    output_location_type: str = "default",
    output_location: Optional[LatchDir] = None,
    conditions_source: str = "manual",
    manual_conditions: Annotated[
        List[List[str]],
        FlyteAnnotation({"_tmp_hack_deseq2": "manual_design_matrix"}),
    ] = [],
    conditions_table: Optional[
        Annotated[
            LatchFile,
            FlyteAnnotation(
                {
                    "_tmp_hack_deseq2": "design_matrix",
                    "rules": [
                        {
                            "regex": r".*\.(csv|tsv|xlsx)$",
                            "message": "Expected a CSV, TSV, or XLSX file",
                        }
                    ],
                }
            ),
        ]
    ] = None,
    design_matrix_sample_id_column: Optional[
        Annotated[str, FlyteAnnotation({"_tmp_hack_deseq2": "design_id_column"})]
    ] = None,
    design_formula: Annotated[
        List[List[str]],
        FlyteAnnotation(
            {
                "_tmp_hack_deseq2": "design_formula",
                "_tmp_hack_deseq2_allow_clustering": True,
            }
        ),
    ] = [],
    number_of_genes_to_plot: int = 30,
) -> LatchDir: ...



from dataclasses import dataclass
from typing import ClassVar, List, Literal, Optional, TypedDict, Union

from typing_extensions import Self, TypeAlias


class InvalidValue(TypedDict):
    rawValue: str
    valid: Literal[False]


class PrimitiveStringValueValid(TypedDict):
    value: str
    valid: Literal[True]


PrimitiveStringValue: TypeAlias = Union[PrimitiveStringValueValid, InvalidValue]


class PrimitiveNumberValueValid(TypedDict):
    value: float
    valid: Literal[True]


PrimitiveNumberValue: TypeAlias = Union[PrimitiveNumberValueValid, InvalidValue]


class PrimitiveNullValueValid(TypedDict):
    value: None
    valid: Literal[True]


PrimitiveNullValue: TypeAlias = Union[PrimitiveNullValueValid, InvalidValue]


class PrimitiveBooleanValueValid(TypedDict):
    value: bool
    valid: Literal[True]


PrimitiveBooleanValue: TypeAlias = Union[PrimitiveBooleanValueValid, InvalidValue]


class BlobValue(TypedDict):
    ldataNodeId: str


class PrimitiveBlobValueValid(TypedDict):
    value: BlobValue
    valid: Literal[True]


PrimitiveBlobValue: TypeAlias = Union[PrimitiveBlobValueValid, InvalidValue]


class LinkValue(TypedDict):
    sampleId: str


class PrimitiveLinkValueValid(TypedDict):
    value: LinkValue
    valid: Literal[True]


PrimitiveLinkValue: TypeAlias = Union[PrimitiveLinkValueValid, InvalidValue]


class PrimitiveEnumValueValid(TypedDict):
    value: str
    valid: Literal[True]


PrimitiveEnumValue: TypeAlias = Union[PrimitiveEnumValueValid, InvalidValue]

PrimitiveValue: TypeAlias = Union[
    PrimitiveStringValue,
    PrimitiveNumberValue,
    PrimitiveNullValue,
    PrimitiveBlobValue,
    PrimitiveLinkValue,
    PrimitiveEnumValue,
    PrimitiveBooleanValue,
]

ArrayValue: TypeAlias = List["DBValue"]


class UnionValue(TypedDict):
    tag: str
    value: "DBValue"


DBValue: TypeAlias = Union[PrimitiveValue, ArrayValue, UnionValue]


@dataclass(frozen=True)
class EmptyCell:
    """Empty Registry :class:`Record` value. Singleton.

    The constructor returns a referentially identical instance each call. That is,
    `EmptyCell() is EmptyCell()`

    Used to distinguish explicit `None` values from missing values.
    """

    _singleton: ClassVar[Optional["EmptyCell"]] = None

    def __new__(cls) -> Self:
        if cls._singleton is None:
            cls._singleton = super().__new__(cls)

        return cls._singleton


Value: TypeAlias = Union[DBValue, EmptyCell]



"""Service to upload test objects to a managed bucket."""

from pathlib import Path

import boto3
import botocore
import click

from latch_cli.services.test_data.utils import _retrieve_creds
from latch_cli.utils import account_id_from_token, retrieve_or_login

BUCKET = "latch-public"


def upload(src_path: str, dont_confirm_overwrite: bool = True) -> str:
    """Uploads a local file/folder to a managed bucket.

    Args:
        src_path: The path of the file/folder to upload.

    Returns:
        s3 URL of uploaded object.

    Example:
        >>> upload("./foo.txt")
            # Puts the file `foo.txt` in a managed bucket and returns a publicly
            # accesible S3 URL.
    """

    session_token, access_key, secret_key, account_id = _retrieve_creds()

    s3_resource = boto3.client(
        "s3",
        aws_access_key_id=access_key,
        aws_secret_access_key=secret_key,
        aws_session_token=session_token,
    )

    if account_id is None or account_id == "":
        account_id = account_id_from_token(retrieve_or_login())

    allowed_key = str((Path("test-data") / account_id).joinpath(src_path))

    upload_helper(
        Path(src_path).resolve(),
        allowed_key,
        s3_resource,
        dont_confirm_overwrite,
    )

    return f"s3://{BUCKET}/{allowed_key}"


def upload_helper(
    src_path: Path,
    key: str,
    s3_resource,
    dont_confirm_overwrite: bool,
):
    if not src_path.exists():
        raise ValueError(f"Path {src_path} doesn't exist.")

    if src_path.is_dir():
        for sub_path in src_path.iterdir():
            upload_helper(
                sub_path, f"{key}/{sub_path.name}", s3_resource, dont_confirm_overwrite
            )
    else:
        upload_file(src_path, key, s3_resource, dont_confirm_overwrite)


def upload_file(
    src_path: Path,
    key: str,
    s3_resource,
    dont_confirm_overwrite: bool,
):
    try:
        s3_resource.head_object(Bucket=BUCKET, Key=key)
    except botocore.exceptions.ClientError:
        pass
    else:
        while True and not dont_confirm_overwrite:
            confirm = input(f"{key} already exists, override it?  (y/n) > ")
            if confirm in ("n", "no"):
                print("Aborting upload.")
                return
            elif confirm in ("y", "yes"):
                break
            else:
                print("Invalid response.")

    s3_resource.upload_file(str(src_path), BUCKET, key)
    click.secho(f"Successfully uploaded {key}", fg="green")
